""" copied from handpose/models/simplepose_QNetLite0.py
"""
# -----------------------------------------------------
# Copyright (c) Shanghai Jiao Tong University. All rights reserved.
# Written by Jiefeng Li (jeff.lee.sjtu@gmail.com)
# -----------------------------------------------------

import torch.nn as nn
from antgo.framework.helper.models.builder import HEADS


@HEADS.register_module()
class SimpleQNetLite0PoseHeatmap2D(nn.Module):
    def __init__(self, num_deconv_filters, num_joints):
        super(SimpleQNetLite0PoseHeatmap2D, self).__init__()
        self.deconv_dim = num_deconv_filters
        self.deconv_layer0, self.deconv_layer1 = self._make_deconv_layer()

        self.fuse_keypoint = nn.Sequential(
            nn.Conv2d(160, 256, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(256),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 160, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(160),
            nn.ReLU(inplace=True),
        )

        self.fuse_up16 = nn.Sequential(
            nn.Conv2d(self.deconv_dim[1], 128, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(128),
            nn.ReLU(inplace=True),
            nn.Conv2d(128, 128, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(128),
            nn.ReLU(inplace=True),
        )

        self.final_layer = nn.Conv2d(
            128, num_joints, kernel_size=3, stride=1, padding=1
        )  # up32 = torch.cat[up32, x32] x32's channel is 32
        self.final_offset_layer = nn.Conv2d(
            128, num_joints * 2, kernel_size=3, stride=1, padding=1
        )  # up32 = torch.cat[up32, x32]  x32's channel is 32

    def _make_deconv_layer(self):
        deconv_layer0 = []
        deconv_layer1 = []
        # deconv_layer2 = []
        # deconv1 = nn.ConvTranspose2d(
        #     2048, self.deconv_dim[0], kernel_size=4, stride=2, padding=int(4 / 2) - 1, bias=False)
        deconv1 = nn.Upsample(size=(8, 8), mode="nearest")  # ,align_corners=False
        # deconv1 = nn.Upsample(scale_factor=2, mode='nearest')
        conv1_1x1 = nn.Conv2d(
            160, self.deconv_dim[0], kernel_size=3, stride=1, padding=1, bias=False
        )  # 2048-->512 for resnet18 @lyp
        bn1 = nn.BatchNorm2d(self.deconv_dim[0])
        # deconv2 = nn.ConvTranspose2d(
        #     self.deconv_dim[0], self.deconv_dim[1], kernel_size=4, stride=2, padding=int(4 / 2) - 1, bias=False)
        deconv2 = nn.Upsample(size=(16, 16), mode="nearest")
        # deconv2 = nn.Upsample(scale_factor=2, mode='nearest')
        conv2_1x1 = nn.Conv2d(self.deconv_dim[0], self.deconv_dim[1], kernel_size=3, stride=1, padding=1, bias=False)
        bn2 = nn.BatchNorm2d(self.deconv_dim[1])

        deconv_layer0.append(deconv1)
        deconv_layer0.append(conv1_1x1)
        deconv_layer0.append(bn1)
        deconv_layer0.append(nn.ReLU(inplace=True))

        deconv_layer1.append(deconv2)
        deconv_layer1.append(conv2_1x1)
        deconv_layer1.append(bn2)
        deconv_layer1.append(nn.ReLU(inplace=True))

        # deconv_layer2.append(deconv3)
        # deconv_layer2.append(conv3_1x1)
        # deconv_layer2.append(bn3)
        # deconv_layer2.append(nn.ReLU(inplace=True))

        return [nn.Sequential(*deconv_layer0), nn.Sequential(*deconv_layer1)]

    def forward(self, *args):
        # x32, x16, x8, x4 = self.preact(im)  # c: 64, 128, 192, 256

        x32, x16, x8, x4 = args[0]
        # his_feat = self.keypoint_layer(his_feat)
        # x4_inv = FTL_invK(x4, K_inv)
        x4_inv = x4
        # x4_fuse = torch.cat([x4, his_feat], dim=1)

        x4_fuse = self.fuse_keypoint(x4_inv)

        up8 = self.deconv_layer0(x4_fuse)

        up16 = self.deconv_layer1(up8)

        up16 = self.fuse_up16(up16)

        uv_heatmap = self.final_layer(up16)
        uv_off = self.final_offset_layer(up16)
        return [uv_heatmap, uv_off]




import random

import numpy as np
import torch


def worker_init_fn(worker_id: int, num_workers: int, rank: int, seed: int):
    """Function to initialize each worker.

    The seed of each worker equals to
    ``num_worker * rank + worker_id + user_seed``.

    Args:
        worker_id (int): Id for each worker.
        num_workers (int): Number of workers.
        rank (int): Rank in distributed training.
        seed (int): Random seed.
    """
    worker_seed = num_workers * rank + worker_id + seed
    np.random.seed(worker_seed)
    random.seed(worker_seed)
    torch.manual_seed(worker_seed)



# Copyright (c) OpenMMLab. All rights reserved.
from antgo.framework.helper.utils import Registry

CONV_LAYERS = Registry('conv layer')
NORM_LAYERS = Registry('norm layer')
ACTIVATION_LAYERS = Registry('activation layer')
PADDING_LAYERS = Registry('padding layer')
UPSAMPLE_LAYERS = Registry('upsample layer')
PLUGIN_LAYERS = Registry('plugin layer')

DROPOUT_LAYERS = Registry('drop out layers')
POSITIONAL_ENCODING = Registry('position encoding')
ATTENTION = Registry('attention')
FEEDFORWARD_NETWORK = Registry('feed-forward Network')
TRANSFORMER_LAYER = Registry('transformerLayer')
TRANSFORMER_LAYER_SEQUENCE = Registry('transformer-layers sequence')



# -*- coding: UTF-8 -*-
# @Time    : 2019/1/22 1:16 PM
# @File    : activelearning.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
# from antgo.resource.html import *
# from antgo.ant.base import *
# from antgo.activelearning.samplingmethods.kcenter_greedy import *
# from antgo.crowdsource.activelearning_server import *
# from antgo.dataflow.common import *
# from antgo.dataflow.recorder import *
# from antvis.client.httprpc import *
# from multiprocessing import Process, Queue
# from antgo.task.task import *
# from scipy.stats import entropy
# import traceback
# import subprocess
# import os
# import socket
# import requests
# import json
# import zipfile

# def _is_open(check_ip, port):
#   s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
#   try:
#     s.connect((check_ip, int(port)))
#     s.shutdown(2)
#     return True
#   except:
#     return False


# def _pick_idle_port(from_port=40000, check_count=100):
#   check_port = from_port
#   while check_count:
#     if not _is_open('127.0.0.1', check_port):
#       break

#     logger.warn('Port %d is occupied, try to use %d port.'%(int(check_port), int(check_port + 1)))

#     check_port += 1
#     check_count -= 1

#     if check_count == 0:
#       check_port = None

#   if check_port is None:
#     logger.warn('Couldnt find valid free port.')
#     exit(-1)

#   return check_port


# class AntActiveLearning(AntBase):
#   def __init__(self,
#                ant_context,
#                ant_name,
#                ant_data_folder,
#                ant_dump_dir,
#                ant_token,
#                ant_task_config=None,
#                **kwargs):
#     super(AntActiveLearning, self).__init__(ant_name, ant_context, ant_token, **kwargs)

#     self.skip_first_training = self.context.params.system['skip_training']
#     self.max_iterators = self.context.params.activelearning.get('max_iterators', 10)
#     if self.max_iterators is None:
#       self.max_iterators = 10

#     self.dump_dir = ant_dump_dir
#     self.web_server_port = self.context.params.system['port']
#     self.web_server_port = int(self.web_server_port) if self.web_server_port is not None else None
#     self.html_template = kwargs.get('html_template', None)
#     self.keywords_template = {}

#     self.ant_task_config = ant_task_config
#     self.ant_data_source = ant_data_folder
#     self.devices = self.context.params.system['devices']
#     self.rpc = None

#   def _core_set_algorithm(self, unlabeled_pool, num):
#     if num > len(unlabeled_pool):
#       logger.warn("Dont have enough unlabeled data(%d), return all."%len(unlabeled_pool))
#       return unlabeled_pool

#     # data: 1XC (C表示特征维度)
#     feature_data = np.array([data['feature'] for data in unlabeled_pool])
#     channels = feature_data.shape[-1]
#     kcg = kCenterGreedy(feature_data.reshape(-1, channels))
#     next_selected = kcg.select_batch(model=None, already_selected=[], N=num)
#     next_selected = [unlabeled_pool[int(s)] for s in next_selected]

#     return next_selected

#   def _entroy_algorithm(self, unlabeled_pool, num):
#     if num > len(unlabeled_pool):
#       logger.warn("Dont have enough unlabeled data(%d), return all."%len(unlabeled_pool))
#       return unlabeled_pool

#     unlabeled_entropy = []
#     for index, data in enumerate(unlabeled_pool):
#       # data: DxC（D表示实验次数，C表示概率信息）
#       assert(len(data['feature'].shape) == 2)
#       p = np.mean(data['feature'], axis=0)
#       unlabeled_entropy.append((entropy(p), index))

#     ordered_unlabeled = sorted(unlabeled_entropy, key=lambda x: x[0], reverse=True)
#     next_selected = [unlabeled_pool[s[1]] for s in ordered_unlabeled[0:num]]
#     return next_selected

#   def _bald_algorithm(self, unlabeled_pool, num):
#     if num > len(unlabeled_pool):
#       logger.warn("Dont have enough unlabeled data(%d), return all."%len(unlabeled_pool))
#       return unlabeled_pool

#     unlabeled_entropy = []
#     for index, data in enumerate(unlabeled_pool):
#       # data: DxC（D表示实验次数，C表示概率信息）
#       assert(len(data['feature'].shape) == 2)
#       part1 = entropy(np.mean(data['feature'], axis=0))

#       experiment_num = data['feature'].shape[0]
#       part2 = []
#       for experiment_i in range(experiment_num):
#         part2.append(entropy(data['feature'][experiment_i:experiment_i+1,:]))
#       part2 = np.mean(part2)
#       unlabeled_entropy.append((part1-part2, index))

#     ordered_unlabeled = sorted(unlabeled_entropy, key=lambda x: x[0], reverse=True)
#     next_selected = [unlabeled_pool[s[1]] for s in ordered_unlabeled[0:num]]
#     return next_selected


#   def _variance_algorithm(self, unlabeled_pool, num):
#     if num > len(unlabeled_pool):
#       logger.warn("Dont have enough unlabeled data(%d), return all."%len(unlabeled_pool))      
#       return unlabeled_pool

#     unlabeled_var = []
#     for index, data in enumerate(unlabeled_pool):
#       # data: DxC (D表示试验次数，C表示概率信息)
#       assert(len(data['feature'].shape) == 2)
#       var_val = np.var(data['feature'], axis=0)
#       unlabeled_var.append(((float)(np.mean(var_val)), index))
    
#     ordered_unlabeled = sorted(unlabeled_var, key=lambda x: x[0], reverse=True)
#     next_selected = [unlabeled_pool[s[1]] for s in ordered_unlabeled[0:num]]
#     return next_selected

#   def _unform_sampling_algorithm(self, unlabeled_pool, num):
#     if num > len(unlabeled_pool):
#       logger.warn("Dont have enough unlabeled data(%d), return all."%len(unlabeled_pool))
#       return unlabeled_pool
    
#     return np.random.choice(unlabeled_pool, num, False)

#   def _waiting_label_sample_select(self, unlabeled_pool, num):
#     sampling_strategy = self.context.params.activelearning.get('sampling_strategy', 'entropy')
#     logger.info('Using %s Strategy to select waiting label data.'%sampling_strategy)
#     if sampling_strategy == 'coreset':
#       return self._core_set_algorithm(unlabeled_pool, num)
#     elif sampling_strategy == 'entropy':
#       return self._entroy_algorithm(unlabeled_pool, num)
#     elif sampling_strategy == 'bald':
#       return self._bald_algorithm(unlabeled_pool, num)
#     elif sampling_strategy == 'variance':
#       return self._variance_algorithm(unlabeled_pool, num)
#     elif sampling_strategy == 'random':
#       return self._unform_sampling_algorithm(unlabeled_pool, num)
#     else:
#       return self._unform_sampling_algorithm(unlabeled_pool, num)

#   def start(self):
#     # 0.step loading challenge task
#     running_ant_task = None
#     if self.token is not None:
#       # 0.step load challenge task
#       response = self.context.dashboard.challenge.get(command=type(self).__name__)
#       if response['status'] is None:
#         # invalid token
#         logger.error('Couldnt load challenge task.')
#         self.token = None
#       elif response['status'] == 'SUSPEND':
#         # prohibit submit challenge task frequently
#         # submit only one in one week
#         logger.error('Prohibit submit challenge task frequently.')
#         exit(-1)
#       elif response['status'] == 'OK':
#         # maybe user token or task token
#         content = response['content']
#         if 'task' in content:
#           challenge_task = create_task_from_json(content)
#           if challenge_task is None:
#             logger.error('Couldnt load challenge task.')
#             exit(-1)
#           running_ant_task = challenge_task
#       else:
#         # unknow error
#         logger.error('Unknow error.')
#         exit(-1)

#     if running_ant_task is None:
#       # 0.step load custom task
#       custom_task = create_task_from_xml(self.ant_task_config, self.context)
#       if custom_task is None:
#         logger.error('Couldnt load custom task.')
#         exit(0)
#       running_ant_task = custom_task

#     assert (running_ant_task is not None)

#     # 配置html模板信息
#     self.keywords_template['TASK_TITLE'] = running_ant_task.task_name
#     self.keywords_template['TASK_TYPE'] = running_ant_task.task_type
#     self.keywords_template['UNLABELED_NUM'] = 0
#     self.keywords_template['LABEL_COMPLEX'] = 0

#     # dataset
#     dataset = \
#         running_ant_task.dataset('train',os.path.join(self.ant_data_source, running_ant_task.dataset_name), running_ant_task.dataset_params)

#     # 如果处于研究状态，自动生成未标注数据（等价于训练集）
#     if self.context.params.system['research']:
#       unlabeled_folder = os.path.join(dataset.dir, dataset.unlabeled_tag)
#       if not os.path.exists(unlabeled_folder):
#         os.makedirs(unlabeled_folder)
        
#         # 生成空文件
#         for index in range(dataset.size):
#           with open('%s/%d'%(unlabeled_folder,index), 'w') as fp:
#               fp.write('')

#     use_local_candidate_file = False
#     if os.path.exists(os.path.join(self.main_folder, 'candidates.txt')):
#       use_local_candidate_file = True
#       logger.info('Using candidate file %s'%os.path.join(self.main_folder, 'candidates.txt'))
#       dataset.candidate_file = os.path.join(self.main_folder, 'candidates.txt')
  
#     # prepare workspace
#     if not os.path.exists(os.path.join(self.main_folder, 'show', 'static', 'data')):
#       os.makedirs(os.path.join(self.main_folder, 'show', 'static', 'data'))

#     annotation_folder = os.path.join(self.main_folder, 'show', 'static', 'data', 'annotations')
#     if not os.path.exists(annotation_folder):
#       os.makedirs(annotation_folder)

#     data_folder = os.path.join(self.main_folder, 'show', 'static', 'data', 'images')
#     if not os.path.exists(data_folder):
#       os.makedirs(data_folder)

#     download_folder = os.path.join(self.main_folder, 'show', 'static', 'data', 'download')
#     if not os.path.exists(download_folder):
#       os.makedirs(download_folder)

#     upload_folder = os.path.join(self.main_folder, 'show', 'static', 'data', 'upload')
#     if not os.path.exists(upload_folder):
#       os.makedirs(upload_folder)

#     # 数据队列
#     request_queue = Queue()

#     # launch show server
#     if self.web_server_port is None:
#       self.web_server_port = 10000
#     self.web_server_port = _pick_idle_port(self.web_server_port)

#     logger.info('Launch active learning show server on port %d.'%self.web_server_port)
#     process = multiprocessing.Process(target=activelearning_web_server,
#                                       args=('activelearning',
#                                             self.main_folder,
#                                             self.html_template,
#                                             self.keywords_template,
#                                             running_ant_task,
#                                             self.web_server_port,
#                                             os.getpid(),
#                                             download_folder,
#                                             upload_folder,
#                                             request_queue))
#     process.daemon = True
#     process.start()
#     logger.info('Waiting 5 seconds for launching show server.')
#     time.sleep(5)

#     self.rpc = HttpRpc('v1','activelearning','127.0.0.1',self.web_server_port)
#     avg_analyze_time = 0

#     # prepare waiting unlabeled data
#     try_iter = 0
#     experiment_id = None
#     while try_iter < self.max_iterators:
#       logger.info('Finding unlabeled and cnadidates size.')
#       unlabeled_dataset_size = dataset.unlabeled_size()
#       labeled_dataset_size = dataset.candidates_size()
      
#       logger.info('Unlabeled data size %d, Labeled data size %d'%(unlabeled_dataset_size, labeled_dataset_size))
#       min_sampling_num = self.context.params.activelearning.get('min_sampling_num', None)
#       if min_sampling_num is not None:
#         min_sampling_num = (int)(min_sampling_num)
#         if unlabeled_dataset_size < min_sampling_num:
#           logger.info('Active learning is over. (unlabeled sample size %d < %d).'%(unlabeled_dataset_size, min_sampling_num))
#           return

#       if unlabeled_dataset_size < 10:
#         logger.info('Active learning is over. (unlabeled sample size %d < 10).'%unlabeled_dataset_size)
#         return

#       logger.info("Round %d, unlabeled dataset size %d, labeled dataset size %d."%(try_iter, unlabeled_dataset_size, labeled_dataset_size))
#       if not os.path.exists(self.dump_dir):
#         os.makedirs(self.dump_dir)

#       # 当前阶段
#       self.stage = "ACTIVELEARNING-TRAIN-ROUND-%d"%try_iter

#       # 通知处理状态(开始未标注数据集挖掘)
#       self.rpc.state.patch(round=try_iter, process_state="UNLABEL-PREPARE")
#       # 开始分析时间
#       analyze_start_time = time.time()
#       if try_iter == 0:
#         experiment_id = self.context.from_experiment

#       if (not self.skip_first_training or try_iter > 0) and labeled_dataset_size > 0:
#         # shell call
#         logger.info('Start training using all labeled data (%d iter).'%try_iter)

#         if os.path.exists(os.path.join(self.dump_dir, 'try_round_train_%d'%try_iter)):
#           shutil.rmtree(os.path.join(self.dump_dir, 'try_round_train_%d'%try_iter))
#         os.makedirs(os.path.join(self.dump_dir, 'try_round_train_%d'%try_iter))

#         cmd_shell = 'antgo train --main_file=%s --main_param=%s' % (self.main_file, self.main_param)
#         cmd_shell += ' --dump=%s/%s' % (self.dump_dir, 'try_round_train_%d'%try_iter)
#         cmd_shell += ' --main_folder=%s' % self.main_folder
#         cmd_shell += ' --task=%s' % self.ant_task_config.split('/')[-1]
#         if experiment_id is not None:
#           cmd_shell += ' --from_experiment=%s' % experiment_id
#         cmd_shell += ' --candidate'
#         cmd_shell += ' --devices=%s' % self.devices
#         cmd_shell += ' --dataset=%s/train' % running_ant_task.dataset_name
#         cmd_shell += ' --name=%s_train_round_%d' % (self.ant_name, try_iter)
#         if use_local_candidate_file:
#           cmd_shell += ' --param=candidate_file:%s/candidates.txt'%self.main_folder
#         training_p = \
#             subprocess.Popen('%s > %s.log' % (cmd_shell, '%s_try_rounnd_train_%d'%(self.name, try_iter)), 
#                               shell=True, 
#                               cwd=self.main_folder)

#         # waiting untile finish training
#         training_p.wait()

#         # 根据返回结果，判断是否正常结束
#         if training_p.returncode != 0:
#           logger.error('Training process exit anomaly.')
#           exit(-1)

#         # 获取训练完成后的实验目录ß
#         experiment_prefix = 'try_round_train_%d'%try_iter
#         experiment_id = None
#         for k in os.listdir(os.path.join(self.dump_dir, experiment_prefix)):
#           if k[0] == '.':
#             continue

#           if os.path.isdir(os.path.join(self.dump_dir, experiment_prefix, k)):
#             experiment_id = os.path.join(self.dump_dir, experiment_prefix, k)
#             break
      
#       # 2.step inference using unlabeled data
#       # 当前阶段
#       self.stage = "ACTIVELEARNING-ANALYZE-ROUND-%d"%try_iter
#       if os.path.exists(os.path.join(self.dump_dir, 'try_round_analyze_%d'%try_iter)):
#         shutil.rmtree(os.path.join(self.dump_dir, 'try_round_analyze_%d'%try_iter))
#       os.makedirs(os.path.join(self.dump_dir, 'try_round_analyze_%d'%try_iter))        

#       logger.info('Start analyze all unlabeled data distribution (%d iter).'%try_iter)
#       cmd_shell = 'antgo predict --main_file=%s --main_param=%s'%(self.main_file, self.main_param)
#       cmd_shell += ' --main_folder=%s' % self.main_folder
#       cmd_shell += ' --dump=%s/%s' % (self.dump_dir, 'try_round_analyze_%d'%try_iter)
#       if experiment_id is not None and experiment_id != '':
#         cmd_shell += ' --from_experiment=%s' % experiment_id
#       cmd_shell += ' --task_t=%s' % running_ant_task.task_type
#       cmd_shell += ' --unlabel'
#       cmd_shell += ' --devices=%s' % self.devices
#       cmd_shell += ' --dataset=%s/train' % running_ant_task.dataset_name
#       cmd_shell += ' --name=%s_predict_round_%d' % (self.ant_name, try_iter)
#       if use_local_candidate_file:
#           cmd_shell += ' --param=candidate_file:%s/candidates.txt'%self.main_folder      
#       inference_p = subprocess.Popen('%s > %s.log' %(cmd_shell, '%s_try_rounnd_analyze_%d'%(self.name, try_iter)), 
#                                       shell=True, 
#                                       cwd=self.main_folder)

#       # waiting untile finish inference
#       inference_p.wait()

#       # 根据返回结果，判断是否正常结束
#       if inference_p.returncode != 0:
#         logger.error('Inference process exit anomaly.')
#         exit(-1)

#       # 获取推断完成后的实验目录
#       inference_experiment_prefix = 'try_round_analyze_%d'%try_iter
#       inference_experiment_id = ''
#       for k in os.listdir(os.path.join(self.dump_dir, inference_experiment_prefix)):
#         if k[0] == '.':
#           continue

#         if os.path.isdir(os.path.join(self.dump_dir, inference_experiment_prefix, k)):
#           inference_experiment_id = k
#           break
      
#       record_reader = RecordReader(os.path.join(self.dump_dir, inference_experiment_prefix, inference_experiment_id, 'record'))
#       unlabeled_pool = []
#       for ss in record_reader.iterate_read('groundtruth', 'predict'):
#         gt, feature = ss
#         unlabeled_pool.append({'file_id': gt['file_id'], 'feature': feature, 'id': gt['id']})

#       select_size = self.context.params.activelearning.get('min_sampling_num', None)
#       if select_size is None:
#         min_sampling_ratio = self.context.params.activelearning.get('min_sampling_ratio', None)
#         if min_sampling_ratio is None:
#           min_sampling_ratio = 0.1
#         select_size = int(len(unlabeled_pool) * min_sampling_ratio)
#         if select_size == 0:
#           select_size = len(unlabeled_pool)

#       if select_size == 0:
#         logger.info('Active learning is over. (selecting size == 0.')
#         return

#       next_selected = self._waiting_label_sample_select(unlabeled_pool, select_size)
#       if len(next_selected) == 0:
#         logger.info("Active learning is over. (selecting size == 0).")
#         return

#       logger.info("Round %d, selecting size %d by %s method."%(try_iter, select_size, self.context.params.activelearning.get('sampling_strategy', 'entropy')))

#       # 结束分析时间
#       analyze_end_time = time.time()

#       # 获得平均分析时间
#       avg_analyze_time = (avg_analyze_time * try_iter + (analyze_end_time - analyze_start_time)) / (try_iter + 1)

#       next_unlabeled_sample_ids = []
#       for f in next_selected:
#         next_unlabeled_sample_ids.append((f['file_id'], f['id']))

#       # 打包等待下一步进行标注的样本
#       tar_file = "round_%d.tar.gz"%try_iter
#       tar_path = os.path.join(download_folder, tar_file)
#       if os.path.exists(tar_path):
#         os.remove(tar_path)

#       tar = tarfile.open(tar_path, "w:gz")
#       for next_unlabeled_sample_id, _ in next_unlabeled_sample_ids:
#         tar.add(os.path.join(dataset.dir, next_unlabeled_sample_id),
#                 arcname="round_%d/%s"%(try_iter,next_unlabeled_sample_id))
#       tar.close()

#       # 研究模式，在研究模式下，标注结果自动获取
#       if self.context.params.system['research']:
#         # 获取标注数据，并打包保存
#         logger.info("Research: auto get label data.")
#         if not os.path.exists(os.path.join(self.dump_dir, 'try_round_auto_label_%d'%try_iter)):
#           os.makedirs(os.path.join(self.dump_dir, 'try_round_auto_label_%d'%try_iter))

#         for file_id, id in next_unlabeled_sample_ids:
#           _, label = dataset.at(id, file_id)
#           label.update({'file_id': file_id, 'id': id})
#           # 去除不可序列化数据
#           filter_label = {}
#           for a, b in label.items():
#             if type(b) != float and type(b) != int and type(b) != list and type(b) != dict and type(b) != tuple and type(b) != str:
#               # 不可序列化数据
#               continue

#             filter_label[a] = b 
          
#           label = filter_label
#           # 自动生成子目录
#           if '/' in file_id:
#             if not os.path.exists(os.path.join(self.dump_dir, 'try_round_auto_label_%d'%try_iter, file_id.split('/')[0])):
#               os.makedirs(os.path.join(self.dump_dir, 'try_round_auto_label_%d'%try_iter, file_id.split('/')[0]))

#           # label 写成文件
#           with open(os.path.join(self.dump_dir, 'try_round_auto_label_%d'%try_iter, file_id), 'w') as fp:
#             json.dump(label, fp)

#         # 使用tar,打包
#         logger.info("Research: warp label data.")
#         tar = tarfile.open(os.path.join(self.dump_dir, 'try_round_auto_label_%d.tar.gz'%try_iter), "w:gz")
#         for file_id, _ in next_unlabeled_sample_ids:
#           tar.add(os.path.join(self.dump_dir, 'try_round_auto_label_%d'%try_iter, file_id),
#                   arcname="try_round_auto_label_%d/%s"%(try_iter, file_id))
#         tar.close()

#         logger.info("Research: copy to %s."%upload_folder)
#         shutil.copy(os.path.join(self.dump_dir, 'try_round_auto_label_%d.tar.gz'%try_iter), upload_folder)
#         request_queue.put({'FILE': 'try_round_auto_label_%d.tar.gz'%try_iter, 'ROUND': try_iter})

#       # 通知处理状态(已经准备好未标注数据集)
#       self.rpc.state.patch(round=try_iter,
#                            process_state="UNLABEL-RESET",
#                            unlabel_dataset=tar_file,
#                            unlabeled_size=unlabeled_dataset_size,
#                            labeled_size=labeled_dataset_size,
#                            round_size=len(next_selected))

#       while True:
#         logger.info('Waiting label (human in loop) in round %d.'%try_iter)
#         request_content = request_queue.get()

#         # 1.step 解压数据集文件
#         logger.info("Untar label dataset.")
#         request_label_dataset = request_content['FILE']
#         request_round = request_content['ROUND']
#         if request_round != try_iter:
#           logger.error('Request label round not consistent.')
#           self.rpc.state.patch(round=try_iter, process_state="LABEL-ERROR")
#           continue

#         # 2.step 建立文件夹
#         if os.path.exists(os.path.join(upload_folder, "round_%d"%request_round)):
#           shutil.rmtree(os.path.join(upload_folder, "round_%d"%request_round))

#         # uncompress
#         if request_label_dataset.endswith('tar') or request_label_dataset.endswith('tar.gz'): 
#           tar = tarfile.open(os.path.join(upload_folder, request_label_dataset))
#           tar_list = tar.getnames()
#           tar.extractall(upload_folder)
#           tar.close()

#           if len(tar_list) == 0:
#             logger.error('Untar file error.')
#             continue

#           if not os.path.exists(os.path.join(upload_folder, "round_%d"%request_round)):
#             # 将解压后的文件夹，修改为 "round_%d"%request_round
#             file_name = os.path.normpath(tar_list[0]).split('/')[0]
#             shutil.move(os.path.join(upload_folder, file_name), os.path.join(upload_folder, "round_%d"%request_round))
#         elif request_label_dataset.endswith('zip'):
#           zip_file = zipfile.ZipFile(os.path.join(upload_folder, request_label_dataset))
#           zip_list = zip_file.namelist()
#           for f in zip_list:
#               zip_file.extract(f, upload_folder)
#           zip_file.close()

#           if len(zip_list) == 0:
#             logger.error('Unzip file error.')
#             continue

#           if not os.path.exists(os.path.join(upload_folder, "round_%d"%request_round)):
#             # 将解压后的文件夹，修改为 "round_%d"%request_round
#             file_name = os.path.normpath(zip_list[0]).split('/')[0]
#             shutil.move(os.path.join(upload_folder, file_name), os.path.join(upload_folder, "round_%d"%request_round))
#         else:
#           logger.error('Dont support upload type.')
#           continue
        
#         # 3.step 检查数据集标准是否符合标准
#         logger.info("Check labeled dataset format.")
#         is_ok = dataset.check_candidate(next_unlabeled_sample_ids, os.path.join(upload_folder, "round_%d"%request_round))
#         if not is_ok:
#           logger.error("Dataset format maybe error, need to update label.")
#           self.rpc.state.patch(round=try_iter, process_state="LABEL-ERROR")
#           continue

#         break

#       self.rpc.state.patch(round=try_iter,
#                            process_state='LABEL-FINISH',
#                            next_round_waiting=avg_analyze_time)

#       logger.info('Prepare candidates.txt by using new labeled data.')
#       for sample_file, sample_id in next_unlabeled_sample_ids:
#         if os.path.exists(os.path.join(upload_folder, 'round_%d'%try_iter, sample_file)):
#           dataset.make_candidate(sample_id,
#                                   sample_file,
#                                   os.path.join(upload_folder, 'round_%d'%try_iter, sample_file),
#                                   'OK')

#       logger.info('Finish round %d label, start next round.'%try_iter)
#       # increment round
#       try_iter += 1



# Copyright (c) 2021  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import paddle
from functools import partial


def add_parambase_function(func):
    setattr(paddle.fluid.framework.ParamBase, func.__name__, func)


@add_parambase_function
def normal_(self, mean=0.0, std=1.0):
    replaced_param = paddle.create_parameter(
        shape=self.shape,
        dtype=self.dtype,
        default_initializer=paddle.nn.initializer.Normal(
            mean=mean, std=std))
    paddle.assign(self, replaced_param)


@add_parambase_function
def zero_(self):
    replaced_param = paddle.create_parameter(
        shape=self.shape,
        dtype=self.dtype,
        default_initializer=paddle.nn.initializer.Assign(
            paddle.zeros(self.shape, self.dtype)))
    paddle.assign(self, replaced_param)


@add_parambase_function
def fill_(self, value):
    replaced_param = paddle.create_parameter(
        shape=self.shape,
        dtype=self.dtype,
        default_initializer=paddle.nn.initializer.Assign(
            paddle.full(self.shape, value, self.dtype)))
    paddle.assign(self, replaced_param)



# -*- coding: UTF-8 -*-
# @Time    : 2021/11/9 10:25 下午
# @File    : proxy_dataset.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import os
from antgo.dataflow.dataset.dataset import *


class ProxyDataset(Dataset):
  def __init__(self, train_or_test, dir=None, params=None):
    super(ProxyDataset, self).__init__(train_or_test, dir, params)
    self.train_data = None
    self.val_data = None
    self.test_data = None

  def split(self, split_params={}, split_method='holdout'):
    assert(self.train_or_test == 'train')
    assert(split_method == 'holdout')
    val_pd = ProxyDataset('val', self.dir, self.ext_params)
    val_pd.register(None, self.val_data, None)
    return self, val_pd

  @property
  def size(self):
    if self.train_or_test == 'train':
      return len(self.train_data)
    elif self.train_or_test == 'val':
      return len(self.val_data)
    else:
      return len(self.test_data)

  def at(self, index):
    if self.train_or_test == 'train':
      return self.train_data[index]
    elif self.train_or_test == 'val':
      return self.val_data[index]
    else:
      return self.test_data[index]

  def data_pool(self):
    for index in range(self.size):
      yield self.at(index)
  
  def register(self, train=None, val=None, test=None):
    self.train_data = train
    self.val_data = val
    self.test_data = test



from .mgd import *
from .ab import *
from .afd import *
from .at import *
from .cc import *
from .crd import *
from .dml import *
from .fitnet import *
from .fsp import *
from .ft import *
from .irg import *
from .logits import *
from .lwm import *
from .nst import *
from .ofd import *
from .pkt import *
from .rkd import *
from .sobolev import *
from .sp import *
from .st import *
from .vid import *
from .hcl import *


from .builder import DATASETS, PIPELINES, build_dataloader, build_dataset, build_kv_dataloader, build_iter_dataloader
from .dataset_wrappers import (ConcatDataset, IterConcatDataset, RepeatDataset)
from .pipelines import *
from .dataset_split import DatasetSamplingByClass
from .dataset_filter import IterableDatasetFilter
from .tfdataset import *
from antgo.dataflow import dataset as local_dataset
from antgo.framework.helper.reader import *
import torchvision


def register_antgo_dataset():
    for dataset_module_name in local_dataset.__all__:
        if dataset_module_name == 'Dataset':
            continue

        dataset_module_reader = \
            type(
                dataset_module_name, 
                (Reader,), 
                {   
                    'name': dataset_module_name,
                    '__doc__': f'{dataset_module_name} reader', 
                    '__init__': lambda self, pipeline=None, weak_pipeline=None, strong_pipeline=None, inputs_def=None, **kwargs: 
                        Reader.__init__(self, getattr(local_dataset, self.name)(**kwargs), pipeline=pipeline, weak_pipeline=weak_pipeline, strong_pipeline=strong_pipeline, inputs_def=inputs_def)
                }
            )
        DATASETS.register_module()(dataset_module_reader)


register_antgo_dataset()


def register_torchvision_dataset():
    needed_torchvision_dataset = ['ImageFolder']
    for dataset_module_name in needed_torchvision_dataset:
        if dataset_module_name.startswith('__'):
            continue

        dataset_module_reader = \
            type(
                dataset_module_name, 
                (TVReader,), 
                {   
                    'name': dataset_module_name,
                    '__doc__': f'{dataset_module_name} reader', 
                    '__init__': lambda self, pipeline=None, weak_pipeline=None, strong_pipeline=None, inputs_def=None, **kwargs: 
                        TVReader.__init__(self, getattr(torchvision.datasets, self.name)(**kwargs), pipeline=pipeline, weak_pipeline=weak_pipeline, strong_pipeline=strong_pipeline, inputs_def=inputs_def)
                }
            )
        DATASETS.register_module()(dataset_module_reader)


register_torchvision_dataset()


__all__ = [
    'DATASETS','build_dataloader','build_dataset','ConcatDataset', 'IterConcatDataset','RepeatDataset', 'TFDataset', 'DatasetSamplingByClass', 'IterableDatasetFilter'
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: timer.py
# Author: jian(jian@mltalker.com)
from __future__ import unicode_literals

from contextlib import contextmanager
import time
from collections import defaultdict
import six
import atexit
import numpy as np
from . import logger


class StatCounter(object):
    """ A simple counter"""
    def __init__(self):
        self.reset()

    def feed(self, v):
        self._values.append(v)

    def reset(self):
        self._values = []

    @property
    def count(self):
        return len(self._values)

    @property
    def average(self):
        assert len(self._values)
        return np.mean(self._values)

    @property
    def sum(self):
        assert len(self._values)
        return np.sum(self._values)

    @property
    def max(self):
        assert len(self._values)
        return max(self._values)

class IterSpeedCounter(object):
    """ To count how often some code gets reached"""
    def __init__(self, print_every, name=None):
        self.cnt = 0
        self.print_every = int(print_every)
        self.name = name if name else 'IterSpeed'

    def reset(self):
        self.start = time.time()

    def __call__(self):
        if self.cnt == 0:
            self.reset()
        self.cnt += 1
        if self.cnt % self.print_every != 0:
            return
        t = time.time() - self.start
        logger.info("{}: {:.2f} sec, {} times, {:.3g} sec/time".format(
            self.name, t, self.cnt, t / self.cnt))

@contextmanager
def timed_operation(msg, log_start=False):
    if log_start:
        logger.info('Start {} ...'.format(msg))
    start = time.time()
    yield
    logger.info('{} Finished, Time={:.2f}sec.'.format(
        msg, time.time() - start))

_TOTAL_TIMER_DATA = defaultdict(StatCounter)

@contextmanager
def total_timer(msg):
    start = time.time()
    yield
    t = time.time() - start
    _TOTAL_TIMER_DATA[msg].feed(t)

def get_elapsed_time(msg):
    if len(_TOTAL_TIMER_DATA) == 0:
        return 0.0

    msg_timer = _TOTAL_TIMER_DATA[msg]
    return msg_timer.sum


import copy
import os
import re

REPLACE_MARK_PATTERN = re.compile(r'\$\s*{\s*(.*?)\s*}')


class CodeGenerator:
    def __init__(self, fname):
        self.fname = fname
        with open(fname) as fin:
            self.code = fin.read()
            self.pit = re.finditer(REPLACE_MARK_PATTERN, self.code)

    def __call__(self, **kwargs):
        code = ""
        i = 0
        for m in self.pit:
            span = m.span()
            code += self.code[i:span[0]]
            i = span[1]
            name = m.groups()[0]
            if name not in kwargs:
                raise Exception('There are some variables which are assigned:\n{} in {}'.format(
                    name, self.fname))
            code += str(kwargs[name])
        code += self.code[i:]
        return code


def gen_code(fname):
    return CodeGenerator(fname)


def get_gen_rel_code(path):
    def gen_rel_code(fname):
        return gen_code(os.path.join(path, fname))
    return gen_rel_code



import os
import json
import copy

class SampleGTTemplate(object):
    def __init__(self) -> None:
        me_file_path = os.path.realpath(__file__)
        parent_folder = os.path.dirname(os.path.dirname(me_file_path))
        with open(os.path.join(parent_folder, 'resource', 'templates', 'sample_gt.json'), 'r') as fp:
            self.sample_info = json.load(fp)
        
        with open(os.path.join(parent_folder, 'resource', 'templates', 'sample_meta.json'), 'r') as fp:
            self.meta_info = json.load(fp)   

    def get(self):
        return copy.deepcopy(self.sample_info)

    def meta(self):
        return copy.deepcopy(self.meta_info)



# Copyright (c) OpenMMLab. All rights reserved.
import warnings
from typing import Callable, Dict, List, Optional, Sequence, Tuple, Union

import numpy as np
from antgo.dataflow.dataset.clsdataset import *
import os


def list_dir_or_file(dir_path,
                        list_dir = True,
                        list_file = True,
                        suffix = None,
                        recursive = False):
    """Scan a directory to find the interested directories or files in
    arbitrary order.

    Note:
        :meth:`list_dir_or_file` returns the path relative to ``dir_path``.

    Args:
        dir_path (str | Path): Path of the directory.
        list_dir (bool): List the directories. Default: True.
        list_file (bool): List the path of files. Default: True.
        suffix (str or tuple[str], optional):  File suffix
            that we are interested in. Default: None.
        recursive (bool): If set to True, recursively scan the
            directory. Default: False.

    Yields:
        Iterable[str]: A relative path to ``dir_path``.
    """
    if list_dir and suffix is not None:
        raise TypeError('`suffix` should be None when `list_dir` is True')

    if (suffix is not None) and not isinstance(suffix, (str, tuple)):
        raise TypeError('`suffix` must be a string or tuple of strings')

    root = dir_path

    def _list_dir_or_file(dir_path, list_dir, list_file, suffix,
                            recursive):
        for entry in os.scandir(dir_path):
            if not entry.name.startswith('.') and entry.is_file():
                rel_path = os.path.relpath(entry.path, root)
                if (suffix is None
                        or rel_path.endswith(suffix)) and list_file:
                    yield rel_path
            elif os.path.isdir(entry.path):
                if list_dir:
                    rel_dir = os.path.relpath(entry.path, root)
                    yield rel_dir
                if recursive:
                    yield from _list_dir_or_file(entry.path, list_dir,
                                                    list_file, suffix,
                                                    recursive)

    return _list_dir_or_file(dir_path, list_dir, list_file, suffix,
                                recursive)


def find_folders(root: str):
    """Find classes by folders under a root.

    Args:
        root (string): root directory of folders

    Returns:
        Tuple[List[str], Dict[str, int]]:

        - folders: The name of sub folders under the root.
        - folder_to_idx: The map from folder name to class idx.
    """
    folders = list(
        list_dir_or_file(
            root,
            list_dir=True,
            list_file=False,
            recursive=False,
        ))
    folders.sort()
    folder_to_idx = {folders[i]: i for i in range(len(folders))}
    return folders, folder_to_idx


def get_samples(root: str, folder_to_idx: Dict[str, int],
                is_valid_file: Callable):
    """Make dataset by walking all images under a root.

    Args:
        root (string): root directory of folders
        folder_to_idx (dict): the map from class name to class idx
        is_valid_file (Callable): A function that takes path of a file
            and check if the file is a valid sample file.

    Returns:
        Tuple[list, set]:

        - samples: a list of tuple where each element is (image, class_idx)
        - empty_folders: The folders don't have any valid files.
    """
    samples = []
    available_classes = set()

    for folder_name in sorted(list(folder_to_idx.keys())):
        _dir = os.path.join(root, folder_name)
        files = list(
            list_dir_or_file(
                _dir,
                list_dir=False,
                list_file=True,
                recursive=True,
            ))
        for file in sorted(list(files)):
            if is_valid_file(file):
                path = os.path.join(folder_name, file)
                item = (path, folder_to_idx[folder_name])
                samples.append(item)
                available_classes.add(folder_name)

    empty_folders = set(folder_to_idx.keys()) - available_classes

    return samples, empty_folders


class CustomClsDataset(ClsDataset):
    """Custom dataset for classification.

    The dataset supports two kinds of annotation format.

    1. An annotation file is provided, and each line indicates a sample:

       The sample files: ::

           data_prefix/
           ├── folder_1
           │   ├── xxx.png
           │   ├── xxy.png
           │   └── ...
           └── folder_2
               ├── 123.png
               ├── nsdf3.png
               └── ...

       The annotation file (the first column is the image path and the second
       column is the index of category): ::

            folder_1/xxx.png 0
            folder_1/xxy.png 1
            folder_2/123.png 5
            folder_2/nsdf3.png 3
            ...

       Please specify the name of categories by the argument ``classes``.

    2. The samples are arranged in the specific way: ::

           data_prefix/
           ├── class_x
           │   ├── xxx.png
           │   ├── xxy.png
           │   └── ...
           │       └── xxz.png
           └── class_y
               ├── 123.png
               ├── nsdf3.png
               ├── ...
               └── asd932_.png

    If the ``ann_file`` is specified, the dataset will be generated by the
    first way, otherwise, try the second way.

    Args:
        data_prefix (str): The path of data directory.
        pipeline (Sequence[dict]): A list of dict, where each element
            represents a operation defined in :mod:`mmcls.datasets.pipelines`.
            Defaults to an empty tuple.
        classes (str | Sequence[str], optional): Specify names of classes.

            - If is string, it should be a file path, and the every line of
              the file is a name of a class.
            - If is a sequence of string, every item is a name of class.
            - If is None, use ``cls.CLASSES`` or the names of sub folders
              (If use the second way to arrange samples).

            Defaults to None.
        ann_file (str, optional): The annotation file. If is string, read
            samples paths from the ann_file. If is None, find samples in
            ``data_prefix``. Defaults to None.
        extensions (Sequence[str]): A sequence of allowed extensions. Defaults
            to ('.jpg', '.jpeg', '.png', '.ppm', '.bmp', '.pgm', '.tif').
        test_mode (bool): In train mode or test mode. It's only a mark and
            won't be used in this class. Defaults to False.
        file_client_args (dict, optional): Arguments to instantiate a
            FileClient. See :class:`mmcv.fileio.FileClient` for details.
            If None, automatically inference from the specified path.
            Defaults to None.
    """

    def __init__(self,
                #  data_prefix: str,
                #  pipeline: Sequence = (),
                #  classes: Union[str, Sequence[str], None] = None,
                #  ann_file: Optional[str] = None,
                #  extensions: Sequence[str] = ('.jpg', '.jpeg', '.png', '.ppm',
                #                               '.bmp', '.pgm', '.tif'),
                #  test_mode: bool = False,
                 
                 train_or_test='train',
                 dir=None,
                 params={
                     'data_prefix': '', 
                     'ann_file': None, 
                     'classes': None, 
                     'extensions': ('.jpg', '.jpeg', '.png', '.ppm', '.bmp', '.pgm', '.tif')}
                 ):
        self.extensions = tuple(set([i.lower() for i in params['extensions']]))
        super().__init__(train_or_test,dir,params)

    def _find_samples(self):
        """find samples from ``data_prefix``."""
        classes, folder_to_idx = find_folders(self.data_prefix)
        samples, empty_classes = get_samples(
            self.data_prefix,
            folder_to_idx,
            is_valid_file=self.is_valid_file
        )

        if len(samples) == 0:
            raise RuntimeError(
                f'Found 0 files in subfolders of: {self.data_prefix}. '
                f'Supported extensions are: {",".join(self.extensions)}')

        if self.CLASSES is not None:
            assert len(self.CLASSES) == len(classes), \
                f"The number of subfolders ({len(classes)}) doesn't match " \
                f'the number of specified classes ({len(self.CLASSES)}). ' \
                'Please check the data folder.'
        else:
            self.CLASSES = classes

        if empty_classes:
            warnings.warn(
                'Found no valid file in the folder '
                f'{", ".join(empty_classes)}. '
                f"Supported extensions are: {', '.join(self.extensions)}",
                UserWarning)

        self.folder_to_idx = folder_to_idx

        return samples

    def load_annotations(self):
        """Load image paths and gt_labels."""
        if self.ann_file is None:
            samples = self._find_samples()
        elif isinstance(self.ann_file, str):
            lines = list_from_file(self.ann_file)
            samples = [x.strip().rsplit(' ', 1) for x in lines]
        else:
            raise TypeError('ann_file must be a str or None')

        data_infos = []
        for filename, gt_label in samples:
            info = {}
            info['image_meta'] = {'filename': filename, 'img_prefix': self.data_prefix}
            info['image_file'] = os.path.join(self.data_prefix,filename)
            if not os.path.exists(info['image_file']):
                info['image_file'] = os.path.join(self.data_prefix, self.train_or_test, filename)            
            
            info['gt_label'] = np.array(gt_label, dtype=np.int64)
            info['category_id'] = np.array(gt_label, dtype=np.int64)
            info['id'] = len(data_infos)
            data_infos.append(info)
        return data_infos

    def is_valid_file(self, filename: str) -> bool:
        """Check if a file is a valid sample."""
        return filename.lower().endswith(self.extensions)



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn
import torch.nn.functional as F

from antgo.framework.helper.utils import TORCH_VERSION, build_from_cfg, digit_version
from .registry import ACTIVATION_LAYERS

for module in [
        nn.ReLU, nn.LeakyReLU, nn.PReLU, nn.RReLU, nn.ReLU6, nn.ELU,
        nn.Sigmoid, nn.Tanh
]:
    ACTIVATION_LAYERS.register_module(module=module)


@ACTIVATION_LAYERS.register_module(name='Clip')
@ACTIVATION_LAYERS.register_module()
class Clamp(nn.Module):
    """Clamp activation layer.

    This activation function is to clamp the feature map value within
    :math:`[min, max]`. More details can be found in ``torch.clamp()``.

    Args:
        min (Number | optional): Lower-bound of the range to be clamped to.
            Default to -1.
        max (Number | optional): Upper-bound of the range to be clamped to.
            Default to 1.
    """

    def __init__(self, min=-1., max=1.):
        super(Clamp, self).__init__()
        self.min = min
        self.max = max

    def forward(self, x):
        """Forward function.

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: Clamped tensor.
        """
        return torch.clamp(x, min=self.min, max=self.max)


class GELU(nn.Module):
    r"""Applies the Gaussian Error Linear Units function:

    .. math::
        \text{GELU}(x) = x * \Phi(x)
    where :math:`\Phi(x)` is the Cumulative Distribution Function for
    Gaussian Distribution.

    Shape:
        - Input: :math:`(N, *)` where `*` means, any number of additional
          dimensions
        - Output: :math:`(N, *)`, same shape as the input

    .. image:: scripts/activation_images/GELU.png

    Examples::

        >>> m = nn.GELU()
        >>> input = torch.randn(2)
        >>> output = m(input)
    """

    def forward(self, input):
        return F.gelu(input)


if (TORCH_VERSION == 'parrots'
        or digit_version(TORCH_VERSION) < digit_version('1.4')):
    ACTIVATION_LAYERS.register_module(module=GELU)
else:
    ACTIVATION_LAYERS.register_module(module=nn.GELU)


def build_activation_layer(cfg):
    """Build activation layer.

    Args:
        cfg (dict): The activation layer config, which should contain:

            - type (str): Layer type.
            - layer args: Args needed to instantiate an activation layer.

    Returns:
        nn.Module: Created activation layer.
    """
    return build_from_cfg(cfg, ACTIVATION_LAYERS)



from typing import Dict

import torch
from torch import nn
from antgo.framework.helper.models.builder import MODELS, build_backbone, build_head, build_neck
from antgo.framework.helper.runner import BaseModule
import numpy as np


@MODELS.register_module()
class KeypointNet(BaseModule):
    def __init__(
        self, backbone, head, neck=None, train_cfg=None, test_cfg=None, init_cfg=None):
        super().__init__(init_cfg=init_cfg)

        self.backbone = build_backbone(backbone)
        self.with_neck = False
        if neck is not None:
            self.neck = build_neck(neck)
            self.with_neck = True
        self.head = build_head(head)
        
        self.ohem = True
        self.cls_criterion = torch.nn.BCEWithLogitsLoss(reduction='none' if self.ohem else 'mean')
        self.criterion = torch.nn.MSELoss(reduction='none' if self.ohem else 'mean')
    
    def _loss(self, pred_heatmap, pred_offset_xy, gt_heatmap, joint_mask, heatmap_mask, offset_x, offset_y):
        hard_weight = 2
        mid_weight = 1
        easy_weight = 0.5

        loss_hm = self.cls_criterion(pred_heatmap, gt_heatmap)

        joint_num = pred_heatmap.shape[1]
        loss_offx = self.criterion(pred_offset_xy[:, :joint_num, :, :].mul(heatmap_mask), offset_x.mul(heatmap_mask))
        loss_offy = self.criterion(pred_offset_xy[:, joint_num:, :, :].mul(heatmap_mask), offset_y.mul(heatmap_mask))

        loss_hm = loss_hm * joint_mask
        loss_offx = loss_offx * joint_mask
        loss_offy = loss_offy * joint_mask

        if self.ohem:
            bs, joint_num = loss_hm.shape[:2]
            loss_hm = torch.reshape(loss_hm, (bs * joint_num, -1))
            loss_offx = torch.reshape(loss_offx, (bs * joint_num, -1))
            loss_offy = torch.reshape(loss_offy, (bs * joint_num, -1))

            hm_items = loss_hm.mean(dim=-1)
            offx_items = loss_offx.mean(dim=-1)
            offy_items = loss_offy.mean(dim=-1)
            sortids = torch.argsort(hm_items)
            topids = sortids[: (bs * joint_num // 3)]
            midids = sortids[(bs * joint_num // 3) : (bs * joint_num // 3 * 2)]
            bottomids = sortids[(bs * joint_num // 3 * 2) :]

            loss_hm = (
                (hm_items[topids] * easy_weight).mean()
                + (hm_items[midids] * mid_weight).mean()
                + (hm_items[bottomids] * hard_weight).mean()
            )
            loss_offx = (
                (offx_items[topids] * easy_weight).mean()
                + (offx_items[midids] * mid_weight).mean()
                + (offx_items[bottomids] * hard_weight).mean()
            )
            loss_offy = (
                (offy_items[topids] * easy_weight).mean()
                + (offy_items[midids] * mid_weight).mean()
                + (offy_items[bottomids] * hard_weight).mean()
            )

        return loss_hm, loss_offx, loss_offy
  
    def _compute_loss_with_heatmap(self, uv_heatmap, uv_off_xy, heatmap, heatmap_weight, offset_x, offset_y, joints_vis) -> Dict[str, torch.Tensor]:
        """compute loss"""
        batch_size = uv_heatmap.shape[0]
        joints_vis = (joints_vis.sum(dim=1) > 0).long()
        joint_mask = joints_vis.reshape(batch_size, 1, 1, 1).repeat(
            1, uv_heatmap.shape[1], uv_heatmap.shape[2], uv_heatmap.shape[3]
        )

        # 2d heatmap + depth loass
        loss_hm, loss_offx, loss_offy = self._loss(
            uv_heatmap, uv_off_xy, heatmap, joint_mask, heatmap_weight, offset_x, offset_y
        )

        outloss = dict()
        outloss["loss_uv_hm"] = loss_hm
        outloss["loss_xy_offset"] = 0.1 * (loss_offx + loss_offy)
        return outloss

    def forward_train(self, image, heatmap, offset_x, offset_y, heatmap_weight, joints_vis, **kwargs):
        output = self.backbone(image)   # x32,x16,x8,x4
        output = self.head(output)      # uv_heatmap, uv_off

        uv_heatmap, uv_off = output
        loss_output = \
            self._compute_loss_with_heatmap(
                uv_heatmap, uv_off, heatmap, heatmap_weight, offset_x, offset_y, joints_vis)
        return loss_output

    def get_max_pred_batch(self, batch_heatmaps, batch_offset_x, batch_offset_y):
        batch_size = batch_heatmaps.shape[0]
        num_joints = batch_heatmaps.shape[1]
        width = batch_heatmaps.shape[3]
        heatmaps_reshaped = batch_heatmaps.reshape((batch_size, num_joints, -1))
        idx = np.argmax(heatmaps_reshaped, 2)
        maxvals = np.max(heatmaps_reshaped, 2)

        maxvals = maxvals.reshape((batch_size, num_joints, 1))
        idx = idx.reshape((batch_size, num_joints, 1))

        preds = np.tile(idx, (1, 1, 2)).astype(np.float32)

        preds[:, :, 0] = (preds[:, :, 0]) % width
        preds[:, :, 1] = np.floor((preds[:, :, 1]) / width)

        for r in range(batch_size):
            for c in range(num_joints):
                dx = batch_offset_x[r, c, int(preds[r, c, 1]), int(preds[r, c, 0])]
                dy = batch_offset_y[r, c, int(preds[r, c, 1]), int(preds[r, c, 0])]
                preds[r, c, 0] += dx
                preds[r, c, 1] += dy

        return preds, maxvals

    def forward_test(self, image, **kwargs):
        image_h, image_w = image.shape[2:]
        output = self.backbone(image)   # x32,x16,x8,x4
        output = self.head(output)      # uv_heatmap, uv_off
  
        uv_heatmap, uv_off = output
        # convert to probability
        uv_heatmap = torch.sigmoid(uv_heatmap)
        heatmap_h, heatmap_w = uv_heatmap.shape[2:]
        joint_num = uv_heatmap.shape[1]
        offset_x, offset_y = uv_off[:, :joint_num, :, :], uv_off[:, joint_num:, :, :]
        preds = uv_heatmap.detach().cpu().numpy()
        offset_x = offset_x.detach().cpu().numpy()
        offset_y = offset_y.detach().cpu().numpy()

        preds, score = self.get_max_pred_batch(preds, offset_x, offset_y)
        pred_gt_hm = np.concatenate([preds * (image_h/heatmap_h), score], axis=2)
        results = {
            'pred_joints2d': [sample_joint2d for sample_joint2d in pred_gt_hm],
        }
        return results

    def onnx_export(self, image):
        output = self.backbone(image)   # x32,x16,x8,x4
        output = self.head(output)      # uv_heatmap, uv_off
  
        return output


# -*- coding: UTF-8 -*-
# @Time    : 2022/5/2 14:55
# @File    : optimizer.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import torch


class SGD(torch.optim.SGD):
  def __init__(self,
               learning_rate=0.001,
               parameters=None,
               weight_decay=None,
               grad_clip=None,
               name=None):
    super().__init__(
      parameters,
      lr=learning_rate.learning_rate if isinstance(learning_rate, LRSchedulerProxy) else learning_rate,
      momentum=0,
      dampening=0,
      weight_decay=0.0 if weight_decay is None else weight_decay,
      nesterov=False
    )

    if isinstance(learning_rate, LRSchedulerProxy):
      learning_rate.set_optimizer(self)

  def clear_grad(self):
    self.zero_grad()

  def clear_gradients(self):
    self.zero_grad()

class Adam(torch.optim.Adam):
  def __init__(self,
               learning_rate=0.001,
               beta1=0.9,
               beta2=0.999,
               epsilon=1e-08,
               parameters=None,
               weight_decay=None,
               grad_clip=None,
               name=None,
               lazy_mode=False):
    super().__init__(parameters,
                     lr=learning_rate.learning_rate if isinstance(learning_rate, LRSchedulerProxy) else learning_rate,
                     betas=(beta1, beta2),
                     eps=epsilon,
                     weight_decay=weight_decay, amsgrad=False)

    if isinstance(learning_rate, LRSchedulerProxy):
      learning_rate.set_optimizer(self)

  def clear_grad(self):
    self.zero_grad()

  def clear_gradients(self):
    self.zero_grad()


class RMSProp(torch.optim.RMSprop):
  def __init__(self,
               learning_rate,
               rho=0.95,
               epsilon=1e-06,
               momentum=0.0,
               centered=False,
               parameters=None,
               weight_decay=None,
               grad_clip=None, name=None):
    super().__init__(parameters,
                     lr=learning_rate.learning_rate if isinstance(learning_rate, LRSchedulerProxy) else learning_rate,
                     alpha=rho,
                     eps=epsilon,
                     weight_decay=0 if weight_decay is None else weight_decay,
                     momentum=momentum,
                     centered=centered)

    if isinstance(learning_rate, LRSchedulerProxy):
      learning_rate.set_optimizer(self)

  def clear_grad(self):
    self.zero_grad()

  def clear_gradients(self):
    self.zero_grad()


class LR(object):
  pass


lr = LR()


class LRSchedulerProxy(object):
  def __init__(self, lr_scheduler_func):
    self.create_lr_scheduler_func = lr_scheduler_func
    self.lr_scheduler = None
    self.learning_rate = 0.0

  def set_optimizer(self, optim):
    self.lr_scheduler = self.create_lr_scheduler_func(optim)

  def __getattr__(self, key):
    if key in self.__dict__:
      return self.__dict__[key]
    return getattr(self.lr_scheduler, key)


def add_optimizer_lr_function(func):
  setattr(lr, func.__name__, func)


class _StepDecay(LRSchedulerProxy):
  def __init__(self, learning_rate, step_size, gamma=0.1, last_epoch=- 1, verbose=False):
    super(_StepDecay, self).__init__(
      lambda x: torch.optim.lr_scheduler.StepLR(x, step_size, gamma=gamma, last_epoch=last_epoch, verbose=verbose)
    )
    self.learning_rate = learning_rate

@add_optimizer_lr_function
def StepDecay(learning_rate, step_size, gamma=0.1, last_epoch=- 1, verbose=False):
    return _StepDecay(learning_rate, step_size, gamma, last_epoch, verbose)


class _CosineAnnealingDecay(LRSchedulerProxy):
  def __init__(self, learning_rate, T_max, eta_min=0, last_epoch=- 1, verbose=False):
    super(_CosineAnnealingDecay, self).__init__(
      lambda x: torch.optim.lr_scheduler.CosineAnnealingLR(x, T_max, eta_min=eta_min, last_epoch=last_epoch, verbose=verbose)
    )
    self.learning_rate = learning_rate

@add_optimizer_lr_function
def CosineAnnealingDecay(learning_rate, T_max, eta_min=0, last_epoch=- 1, verbose=False):
    return _CosineAnnealingDecay(learning_rate, T_max, eta_min, last_epoch, verbose)


class _MultiStepDecay(LRSchedulerProxy):
  def __init__(self,learning_rate, milestones, gamma=0.1, last_epoch=- 1, verbose=False):
    super(_MultiStepDecay, self).__init__(
      lambda x: torch.optim.lr_scheduler.MultiStepLR(x, milestones, gamma=gamma, last_epoch=last_epoch, verbose=verbose)
    )
    self.learning_rate = learning_rate

@add_optimizer_lr_function
def MultiStepDecay(learning_rate, milestones, gamma=0.1, last_epoch=- 1, verbose=False):
    return _MultiStepDecay(learning_rate, milestones, gamma, last_epoch, verbose)






# -*- coding: UTF-8 -*-
# @Time    : 17-12-27
# @File    : cityscape.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import sys
import os
import numpy as np
from antgo.dataflow.dataset import *
from antgo.framework.helper.fileio.file_client import *
import cv2
import time


__all__ = ['Cityscape']
class Cityscape(Dataset):
    def __init__(self, train_or_test, dir=None, ext_params=None):
        super(Cityscape, self).__init__(train_or_test, dir, ext_params=ext_params)
        assert(train_or_test in ['train', 'val', 'test'])
        if os.environ.get('LOCAL_RANK', 0) == 0:
            if not os.path.exists(os.path.join(self.dir, 'cityscapes')):
                # 数据集不存在，需要重新下载，并创建标记                
                ali = AliBackend()
                ali.download('ali:///dataset/cityscapes/cityscapes.tar', self.dir)
                os.system(f'cd {self.dir} && tar -xf cityscapes.tar')
                os.system('touch DATASET_IS_READY')
            else:
                # 数据集存在，创建标记
                if not os.path.exists('DATASET_IS_READY'):
                    os.system('touch DATASET_IS_READY')  
        else:
            while True:
                # 等待直到存在指定文件
                if os.path.exists('DATASET_IS_READY'):
                    break
                time.sleep(5)

        self._image_file_list = []
        self._annotation_file_list = []
        index_file = f'{self.train_or_test}.list'
        with open(os.path.join(self.dir, 'cityscapes', index_file), 'r') as fp:
            content = fp.readline()
            content = content.strip()
            
            while content:
                image_file, label_file = content.split(' ')
                label_file = label_file[:-24]+'gtFine_labelIds.png'
                self._image_file_list.append(os.path.join(self.dir, 'cityscapes', image_file))
                self._annotation_file_list.append(os.path.join(self.dir, 'cityscapes', label_file))
                content = fp.readline()
                content = content.strip()
                if content == '':
                    break

    @property
    def size(self):
        return len(self._image_file_list)
        
    def at(self, id):
        image = cv2.imread(self._image_file_list[id])
        segments = cv2.imread(self._annotation_file_list[id], cv2.IMREAD_GRAYSCALE)

        return (
            image, 
            {
                'segments':segments, 
                'image_meta': {
                    'image_shape': (image.shape[0], image.shape[1]),
                    'image_file': self._image_file_list[id]
                }
            }
        )

# cityscape = Cityscape('train', '/root/workspace/dataset/cityscapes')
# size = cityscape.size
# print(f'cityscape size {size}')
# label_num_map = {}
# for i in range(size):
#   data = cityscape.sample(i)
#   ll = set(data['segments'].flatten().tolist())
#   for l in ll:
#     if l not in label_num_map:
#       label_num_map[l] = 0
#     label_num_map[l] += 1

# print(label_num_map)


import os
import sys
import copy
from typing import Any
sys.path.append('.3rd/eagleeye-py/libs/X86-64')


import eagleeye
import numpy as np


class CoreOp(object):
    def __init__(self, func_op_name, **kwargs):
        if '_' in func_op_name:
            a,b = func_op_name.split('_')
            func_op_name = f'{a.capitalize()}{b.capitalize()}'
        self.func_op_name = func_op_name
        
        self.param_1 = dict()   # {"key": [float,float,float,...]}
        self.param_2 = dict()   # {"key": ["","","",...]}
        self.param_3 = dict()   # {"key": [[float,float,...],[],...]}

        for var_key, var_value in kwargs.items():
            if isinstance(var_value, list):
                if len(var_value) > 0:
                    if isinstance(var_value[0], str):
                        self.param_2[var_key] = var_value
                    elif isinstance(var_value[0], list):
                        temp = np.array(var_value).astype(np.float32).tolist()
                        if len(temp.shape) == 1:
                            self.param_3[var_key] = temp
                    else:
                        self.param_1[var_key] = np.array(var_value).astype(np.float32).tolist()
            elif isinstance(var_value, np.ndarray):
                if len(var_value.shape) == 1:
                    self.param_1[var_key] = var_value.astype(np.float32).tolist()
                elif len(var_value.shape) == 2:
                    self.param_3[var_key] = var_value.astype(np.float32).tolist()
                else:
                    print(f'Dont support {var_key}')
                    print(var_value)

    def __call__(self, *args):
        input_tensors = []
        for tensor in args:
            assert(isinstance(tensor, np.ndarray))
            input_tensors.append(tensor)

        output_tensors = eagleeye.execute(self.func_op_name, self.func_op_name,self.param_1, self.param_2,self.param_3, input_tensors)
        return output_tensors if len(output_tensors) > 1 else output_tensors[0]


# Copyright (c) 2021  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import paddle
import PIL
import numbers
import numpy as np
from PIL import Image
from paddle.vision.transforms import BaseTransform
from paddle.vision.transforms import functional as F
from paddle.vision.transforms import *


class ToPILImage(BaseTransform):
    def __init__(self, mode=None, keys=None):
        super(ToPILImage, self).__init__(keys)
        self.data_format = data_format

    def _apply_image(self, pic):
        """
        Args:
            pic (Tensor|np.ndarray): Image to be converted to PIL Image.
        Returns:
            PIL: Converted image.
        """
        if not (isinstance(pic, paddle.Tensor) or isinstance(pic, np.ndarray)):
            raise TypeError('pic should be Tensor or ndarray. Got {}.'.format(
                type(pic)))

        elif isinstance(pic, paddle.Tensor):
            if pic.ndimension() not in {2, 3}:
                raise ValueError(
                    'pic should be 2/3 dimensional. Got {} dimensions.'.format(
                        pic.ndimension()))

            elif pic.ndimension() == 2:
                # if 2D image, add channel dimension (CHW)
                pic = pic.unsqueeze(0)

        elif isinstance(pic, np.ndarray):
            if pic.ndim not in {2, 3}:
                raise ValueError(
                    'pic should be 2/3 dimensional. Got {} dimensions.'.format(
                        pic.ndim))

            elif pic.ndim == 2:
                # if 2D image, add channel dimension (HWC)
                pic = np.expand_dims(pic, 2)

        npimg = pic
        if isinstance(pic, paddle.Tensor) and "float" in str(pic.numpy(
        ).dtype) and mode != 'F':
            pic = pic.mul(255).byte()
        if isinstance(pic, paddle.Tensor):
            npimg = np.transpose(pic.numpy(), (1, 2, 0))

        if not isinstance(npimg, np.ndarray):
            raise TypeError(
                'Input pic must be a paddle.Tensor or NumPy ndarray, ' +
                'not {}'.format(type(npimg)))

        if npimg.shape[2] == 1:
            expected_mode = None
            npimg = npimg[:, :, 0]
            if npimg.dtype == np.uint8:
                expected_mode = 'L'
            elif npimg.dtype == np.int16:
                expected_mode = 'I;16'
            elif npimg.dtype == np.int32:
                expected_mode = 'I'
            elif npimg.dtype == np.float32:
                expected_mode = 'F'
            if mode is not None and mode != expected_mode:
                raise ValueError(
                    "Incorrect mode ({}) supplied for input type {}. Should be {}"
                    .format(mode, np.dtype, expected_mode))
            mode = expected_mode

        elif npimg.shape[2] == 2:
            permitted_2_channel_modes = ['LA']
            if mode is not None and mode not in permitted_2_channel_modes:
                raise ValueError("Only modes {} are supported for 2D inputs".
                                 format(permitted_2_channel_modes))

            if mode is None and npimg.dtype == np.uint8:
                mode = 'LA'

        elif npimg.shape[2] == 4:
            permitted_4_channel_modes = ['RGBA', 'CMYK', 'RGBX']
            if mode is not None and mode not in permitted_4_channel_modes:
                raise ValueError("Only modes {} are supported for 4D inputs".
                                 format(permitted_4_channel_modes))

            if mode is None and npimg.dtype == np.uint8:
                mode = 'RGBA'
        else:
            permitted_3_channel_modes = ['RGB', 'YCbCr', 'HSV']
            if mode is not None and mode not in permitted_3_channel_modes:
                raise ValueError("Only modes {} are supported for 3D inputs".
                                 format(permitted_3_channel_modes))
            if mode is None and npimg.dtype == np.uint8:
                mode = 'RGB'

        if mode is None:
            raise TypeError('Input type {} is not supported'.format(
                npimg.dtype))

        return Image.fromarray(npimg, mode=mode)


class ToTensor(BaseTransform):
    """Convert a ``PIL.Image`` or ``numpy.ndarray`` to ``numpy.ndarray`` with shapr (C x H x W).
    Args:
        data_format (str, optional): Data format of output tensor, should be 'HWC' or
            'CHW'. Default: 'CHW'.
        keys (list[str]|tuple[str], optional): Same as ``BaseTransform``. Default: None.
    """

    def __init__(self, data_format='CHW', keys=None):
        super(ToTensor, self).__init__(keys)
        self.data_format = data_format

    def _apply_image(self, img):
        """
        Args:
            img (PIL.Image|np.ndarray): Image to be converted to tensor.
        Returns:
            np.ndarray: Converted image.
        """
        try:
            if isinstance(img, PIL.JpegImagePlugin.JpegImageFile) or isinstance(
                    img, PIL.Image.Image):
                img = np.array(img)
        except:
            pass

        img = img / 255.0
        img = img.transpose((2, 0, 1)).astype("float32")
        img = paddle.to_tensor(img)
        return img


class Normalize(BaseTransform):
    """Normalize the input data with mean and standard deviation.
    Given mean: ``(M1,...,Mn)`` and std: ``(S1,..,Sn)`` for ``n`` channels,
    this transform will normalize each channel of the input data.
    ``output[channel] = (input[channel] - mean[channel]) / std[channel]``
    Args:
        mean (int|float|list): Sequence of means for each channel.
        std (int|float|list): Sequence of standard deviations for each channel.
    """

    def __init__(self, mean=0.0, std=1.0, inplace=False):
        key = None
        super(Normalize, self).__init__(key)
        if isinstance(mean, numbers.Number):
            mean = [mean, mean, mean]

        if isinstance(std, numbers.Number):
            std = [std, std, std]

        self.mean = mean
        self.std = std

    def _apply_image(self, img):
        if isinstance(img, paddle.Tensor):
            img = img.numpy()
        return F.normalize(img, self.mean, self.std, 'CHW', False)


class Lambda(BaseTransform):
    """Apply a user-defined lambda as a transform. This transform does not support torchscript.
    Args:
        lambd (function): Lambda/function to be used for transform.
    """

    def __init__(self, lambd):
        if not callable(lambd):
            raise TypeError("Argument lambd should be callable, got {}".format(
                repr(type(lambd).__name__)))
        self.lambd = lambd

    def _apply_image(self, img):
        return self.lambd(img)


# Copyright (c) OpenMMLab. All rights reserved.
import torch

from ..builder import BBOX_ASSIGNERS
from ..iou_calculators import build_iou_calculator
from .assign_result import AssignResult
from .base_assigner import BaseAssigner


@BBOX_ASSIGNERS.register_module()
class MaxIoUAssigner(BaseAssigner):
    """Assign a corresponding gt bbox or background to each bbox.

    Each proposals will be assigned with `-1`, or a semi-positive integer
    indicating the ground truth index.

    - -1: negative sample, no assigned gt
    - semi-positive integer: positive sample, index (0-based) of assigned gt

    Args:
        pos_iou_thr (float): IoU threshold for positive bboxes.
        neg_iou_thr (float or tuple): IoU threshold for negative bboxes.
        min_pos_iou (float): Minimum iou for a bbox to be considered as a
            positive bbox. Positive samples can have smaller IoU than
            pos_iou_thr due to the 4th step (assign max IoU sample to each gt).
        gt_max_assign_all (bool): Whether to assign all bboxes with the same
            highest overlap with some gt to that gt.
        ignore_iof_thr (float): IoF threshold for ignoring bboxes (if
            `gt_bboxes_ignore` is specified). Negative values mean not
            ignoring any bboxes.
        ignore_wrt_candidates (bool): Whether to compute the iof between
            `bboxes` and `gt_bboxes_ignore`, or the contrary.
        match_low_quality (bool): Whether to allow low quality matches. This is
            usually allowed for RPN and single stage detectors, but not allowed
            in the second stage. Details are demonstrated in Step 4.
        gpu_assign_thr (int): The upper bound of the number of GT for GPU
            assign. When the number of gt is above this threshold, will assign
            on CPU device. Negative values mean not assign on CPU.
    """

    def __init__(self,
                 pos_iou_thr,
                 neg_iou_thr,
                 min_pos_iou=.0,
                 gt_max_assign_all=True,
                 ignore_iof_thr=-1,
                 ignore_wrt_candidates=True,
                 match_low_quality=True,
                 gpu_assign_thr=-1,
                 iou_calculator=dict(type='BboxOverlaps2D')):
        self.pos_iou_thr = pos_iou_thr
        self.neg_iou_thr = neg_iou_thr
        self.min_pos_iou = min_pos_iou
        self.gt_max_assign_all = gt_max_assign_all
        self.ignore_iof_thr = ignore_iof_thr
        self.ignore_wrt_candidates = ignore_wrt_candidates
        self.gpu_assign_thr = gpu_assign_thr
        self.match_low_quality = match_low_quality
        self.iou_calculator = build_iou_calculator(iou_calculator)

    def assign(self, bboxes, gt_bboxes, gt_bboxes_ignore=None, gt_labels=None):
        """Assign gt to bboxes.

        This method assign a gt bbox to every bbox (proposal/anchor), each bbox
        will be assigned with -1, or a semi-positive number. -1 means negative
        sample, semi-positive number is the index (0-based) of assigned gt.
        The assignment is done in following steps, the order matters.

        1. assign every bbox to the background
        2. assign proposals whose iou with all gts < neg_iou_thr to 0
        3. for each bbox, if the iou with its nearest gt >= pos_iou_thr,
           assign it to that bbox
        4. for each gt bbox, assign its nearest proposals (may be more than
           one) to itself

        Args:
            bboxes (Tensor): Bounding boxes to be assigned, shape(n, 4).
            gt_bboxes (Tensor): Groundtruth boxes, shape (k, 4).
            gt_bboxes_ignore (Tensor, optional): Ground truth bboxes that are
                labelled as `ignored`, e.g., crowd boxes in COCO.
            gt_labels (Tensor, optional): Label of gt_bboxes, shape (k, ).

        Returns:
            :obj:`AssignResult`: The assign result.

        Example:
            >>> self = MaxIoUAssigner(0.5, 0.5)
            >>> bboxes = torch.Tensor([[0, 0, 10, 10], [10, 10, 20, 20]])
            >>> gt_bboxes = torch.Tensor([[0, 0, 10, 9]])
            >>> assign_result = self.assign(bboxes, gt_bboxes)
            >>> expected_gt_inds = torch.LongTensor([1, 0])
            >>> assert torch.all(assign_result.gt_inds == expected_gt_inds)
        """
        assign_on_cpu = True if (self.gpu_assign_thr > 0) and (
            gt_bboxes.shape[0] > self.gpu_assign_thr) else False
        # compute overlap and assign gt on CPU when number of GT is large
        if assign_on_cpu:
            device = bboxes.device
            bboxes = bboxes.cpu()
            gt_bboxes = gt_bboxes.cpu()
            if gt_bboxes_ignore is not None:
                gt_bboxes_ignore = gt_bboxes_ignore.cpu()
            if gt_labels is not None:
                gt_labels = gt_labels.cpu()

        overlaps = self.iou_calculator(gt_bboxes, bboxes)

        if (self.ignore_iof_thr > 0 and gt_bboxes_ignore is not None
                and gt_bboxes_ignore.numel() > 0 and bboxes.numel() > 0):
            if self.ignore_wrt_candidates:
                ignore_overlaps = self.iou_calculator(
                    bboxes, gt_bboxes_ignore, mode='iof')
                ignore_max_overlaps, _ = ignore_overlaps.max(dim=1)
            else:
                ignore_overlaps = self.iou_calculator(
                    gt_bboxes_ignore, bboxes, mode='iof')
                ignore_max_overlaps, _ = ignore_overlaps.max(dim=0)
            overlaps[:, ignore_max_overlaps > self.ignore_iof_thr] = -1

        assign_result = self.assign_wrt_overlaps(overlaps, gt_labels)
        if assign_on_cpu:
            assign_result.gt_inds = assign_result.gt_inds.to(device)
            assign_result.max_overlaps = assign_result.max_overlaps.to(device)
            if assign_result.labels is not None:
                assign_result.labels = assign_result.labels.to(device)
        return assign_result

    def assign_wrt_overlaps(self, overlaps, gt_labels=None):
        """Assign w.r.t. the overlaps of bboxes with gts.

        Args:
            overlaps (Tensor): Overlaps between k gt_bboxes and n bboxes,
                shape(k, n).
            gt_labels (Tensor, optional): Labels of k gt_bboxes, shape (k, ).

        Returns:
            :obj:`AssignResult`: The assign result.
        """
        num_gts, num_bboxes = overlaps.size(0), overlaps.size(1)

        # 1. assign -1 by default
        assigned_gt_inds = overlaps.new_full((num_bboxes, ),
                                             -1,
                                             dtype=torch.long)

        if num_gts == 0 or num_bboxes == 0:
            # No ground truth or boxes, return empty assignment
            max_overlaps = overlaps.new_zeros((num_bboxes, ))
            if num_gts == 0:
                # No truth, assign everything to background
                assigned_gt_inds[:] = 0
            if gt_labels is None:
                assigned_labels = None
            else:
                assigned_labels = overlaps.new_full((num_bboxes, ),
                                                    -1,
                                                    dtype=torch.long)
            return AssignResult(
                num_gts,
                assigned_gt_inds,
                max_overlaps,
                labels=assigned_labels)

        # for each anchor, which gt best overlaps with it
        # for each anchor, the max iou of all gts
        max_overlaps, argmax_overlaps = overlaps.max(dim=0)
        # for each gt, which anchor best overlaps with it
        # for each gt, the max iou of all proposals
        gt_max_overlaps, gt_argmax_overlaps = overlaps.max(dim=1)

        # 2. assign negative: below
        # the negative inds are set to be 0
        if isinstance(self.neg_iou_thr, float):
            assigned_gt_inds[(max_overlaps >= 0)
                             & (max_overlaps < self.neg_iou_thr)] = 0
        elif isinstance(self.neg_iou_thr, tuple):
            assert len(self.neg_iou_thr) == 2
            assigned_gt_inds[(max_overlaps >= self.neg_iou_thr[0])
                             & (max_overlaps < self.neg_iou_thr[1])] = 0

        # 3. assign positive: above positive IoU threshold
        pos_inds = max_overlaps >= self.pos_iou_thr
        assigned_gt_inds[pos_inds] = argmax_overlaps[pos_inds] + 1

        if self.match_low_quality:
            # Low-quality matching will overwrite the assigned_gt_inds assigned
            # in Step 3. Thus, the assigned gt might not be the best one for
            # prediction.
            # For example, if bbox A has 0.9 and 0.8 iou with GT bbox 1 & 2,
            # bbox 1 will be assigned as the best target for bbox A in step 3.
            # However, if GT bbox 2's gt_argmax_overlaps = A, bbox A's
            # assigned_gt_inds will be overwritten to be bbox B.
            # This might be the reason that it is not used in ROI Heads.
            for i in range(num_gts):
                if gt_max_overlaps[i] >= self.min_pos_iou:
                    if self.gt_max_assign_all:
                        max_iou_inds = overlaps[i, :] == gt_max_overlaps[i]
                        assigned_gt_inds[max_iou_inds] = i + 1
                    else:
                        assigned_gt_inds[gt_argmax_overlaps[i]] = i + 1

        if gt_labels is not None:
            assigned_labels = assigned_gt_inds.new_full((num_bboxes, ), -1)
            pos_inds = torch.nonzero(
                assigned_gt_inds > 0, as_tuple=False).squeeze()
            if pos_inds.numel() > 0:
                assigned_labels[pos_inds] = gt_labels[
                    assigned_gt_inds[pos_inds] - 1]
        else:
            assigned_labels = None

        return AssignResult(
            num_gts, assigned_gt_inds, max_overlaps, labels=assigned_labels)



# -*- coding: UTF-8 -*-
# @Time : 29/03/2018
# @File : lip.py
# @Author: Jian <jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import sys
from antgo.dataflow.dataset import *
import os
import numpy as np
import cv2
import time
from antgo.framework.helper.fileio.file_client import *


__all__ = ['LIP']
class LIP(Dataset):
  def __init__(self, train_or_test, dir=None, ext_params=None):
    super(LIP, self).__init__(train_or_test, dir, ext_params)
    assert (train_or_test in ['train', 'val'])

    if os.environ.get('LOCAL_RANK', 0) == 0:
      if not os.path.exists(os.path.join(self.dir, 'train_id.txt')):
        # 数据集不存在，需要重新下载，并创建标记
        if not os.path.exists(self.dir):
          os.makedirs(self.dir)

        ali = AliBackend()
        ali.download('ali:///dataset/lip/TrainVal_images.zip', self.dir)
        ali.download('ali:///dataset/lip/Testing_images.zip', self.dir)
        ali.download('ali:///dataset/lip/TrainVal_parsing_annotations.zip', self.dir)
        ali.download('ali:///dataset/lip/train_id.txt', self.dir)
        ali.download('ali:///dataset/lip/val_id.txt', self.dir)

        os.system(f'cd {self.dir} && unzip TrainVal_images.zip && unzip Testing_images.zip && unzip TrainVal_parsing_annotations.zip')
        os.system('touch DATASET_IS_READY')
      else:
        # 数据集存在，创建标记
        if not os.path.exists('DATASET_IS_READY'):
          os.system('touch DATASET_IS_READY')
    else:
      while True:
        # 等待直到存在指定文件
        if os.path.exists('DATASET_IS_READY'):
          break
        time.sleep(5)

    self.data_list = []    
    parse_file = os.path.join(self.dir, '%s' % ('train_id.txt' if self.train_or_test == 'train' else 'val_id.txt'))
    image_folder = os.path.join(self.dir, '%s_images' % ('train' if self.train_or_test == 'train' else 'val'))
    label_folder = os.path.join(self.dir, '%s_segmentations' % ('train' if self.train_or_test == 'train' else 'val'))
    with open(parse_file, 'r') as fp:
      content = fp.readline()
      while content:
        content = content.replace('\n', '')
        self.data_list.append((os.path.join(image_folder, f'{content}.jpg'),
                               os.path.join(label_folder, f'{content}.png')))

        content = fp.readline()

    self.ids = list(range(len(self.data_list)))

    self.inv_category_map = {
      0: 'background',
      1: 'Hat',
      2: 'Hair',
      3: 'Sunglasses',
      4: 'Upper-clothes',
      5: 'Dress',
      6: 'Coat',
      7: 'Socks',
      8: 'Pants',
      9: 'Glove',
      10: 'Scarf',
      11: 'Skirt',
      12: 'Jumpsuits',
      13: 'Face',
      14: 'Right-arm',
      15: 'Left-arm',
      16: 'Right-leg',
      17: 'Left-leg',
      18: 'Right-shoe',
      19: 'Left-shoe'
    }
    self.category_map = {v: k for k, v in self.inv_category_map.items()}

  @property
  def size(self):
    return len(self.ids)

  def at(self, id):
    image_file, label_file = self.data_list[id]
    image = cv2.imread(image_file)
    label = cv2.imread(label_file, cv2.IMREAD_GRAYSCALE)
    return (
      image,
      {
        'segments': label,
        'image_meta': {
            'image_shape': (image.shape[0], image.shape[1]),
            'image_file': image_file
          }   
      }
    )

  def split(self, split_params={}, split_method='holdout'):
    assert (self.train_or_test == 'train')
    assert (split_method == 'holdout')
    val_dataset = LIP('val', self.dir)

    return self, val_dataset

# p2012 = LIP('train', '/root/workspace/dataset/B')
# print(f'p2012 size {p2012.size}')
# for i in range(p2012.size):
#   result = p2012.sample(i)
#   ss = result['segments']
#   # cv2.imwrite('./1234.png', (ss/20*255).astype(np.uint8))
#   print(i)
# value = p2012.sample(0)
# print(value.keys())
# value = p2012.sample(1)
# print(value)


from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import torch
import torch.nn as nn
import torch.nn.functional as F
from ...builder import DISTILL_LOSSES


'''
From https://github.com/lenscloth/RKD/blob/master/metric/loss.py
'''


@DISTILL_LOSSES.register_module()
class RKD(nn.Module):
	'''
	Relational Knowledge Distillation
	https://arxiv.org/pdf/1904.05068.pdf
	'''
	def __init__(self, w_dist, w_angle):
		super(RKD, self).__init__()

		self.w_dist  = w_dist
		self.w_angle = w_angle

	def forward(self, feat_s, feat_t):
		loss = self.w_dist * self.rkd_dist(feat_s, feat_t) + \
			   self.w_angle * self.rkd_angle(feat_s, feat_t)

		return loss

	def rkd_dist(self, feat_s, feat_t):
		feat_t_dist = self.pdist(feat_t, squared=False)
		mean_feat_t_dist = feat_t_dist[feat_t_dist>0].mean()
		feat_t_dist = feat_t_dist / mean_feat_t_dist

		feat_s_dist = self.pdist(feat_s, squared=False)
		mean_feat_s_dist = feat_s_dist[feat_s_dist>0].mean()
		feat_s_dist = feat_s_dist / mean_feat_s_dist

		loss = F.smooth_l1_loss(feat_s_dist, feat_t_dist)

		return loss

	def rkd_angle(self, feat_s, feat_t):
		# N x C --> N x N x C
		feat_t_vd = (feat_t.unsqueeze(0) - feat_t.unsqueeze(1))
		norm_feat_t_vd = F.normalize(feat_t_vd, p=2, dim=2)
		feat_t_angle = torch.bmm(norm_feat_t_vd, norm_feat_t_vd.transpose(1, 2)).view(-1)

		feat_s_vd = (feat_s.unsqueeze(0) - feat_s.unsqueeze(1))
		norm_feat_s_vd = F.normalize(feat_s_vd, p=2, dim=2)
		feat_s_angle = torch.bmm(norm_feat_s_vd, norm_feat_s_vd.transpose(1, 2)).view(-1)

		loss = F.smooth_l1_loss(feat_s_angle, feat_t_angle)

		return loss

	def pdist(self, feat, squared=False, eps=1e-12):
		feat_square = feat.pow(2).sum(dim=1)
		feat_prod   = torch.mm(feat, feat.t())
		feat_dist   = (feat_square.unsqueeze(0) + feat_square.unsqueeze(1) - 2 * feat_prod).clamp(min=eps)

		if not squared:
			feat_dist = feat_dist.sqrt()

		feat_dist = feat_dist.clone()
		feat_dist[range(len(feat)), range(len(feat))] = 0

		return feat_dist



# -*- coding: UTF-8 -*-
# @Time    : 17-12-27
# @File    : vggface.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import sys
import os
import numpy as np
import time
from antgo.dataflow.dataset import *
from antgo.framework.helper.fileio.file_client import *
import cv2


class VGGFace(Dataset):
  def __init__(self, train_or_test, dir=None, ext_params=None):
    if train_or_test != 'train':
      train_or_test = 'test'
    super(VGGFace, self).__init__(train_or_test, dir, ext_params=ext_params)
    assert(train_or_test in ['train', 'test', 'val'])

    if os.environ.get('LOCAL_RANK', 0) == 0:
      if not os.path.exists(os.path.join(self.dir, 'data')) or not os.path.exists(os.path.join(self.dir, 'meta')):
        # 数据集不存在，需要重新下载，并创建标记
        ali = AliBackend()
        ali.download('ali:///dataset/vgg-face2/data', self.dir)
        ali.download('ali:///dataset/vgg-face2/meta', self.dir)

        os.system(f'cd {os.path.join(self.dir, "data")} && tar -xf vggface2_train.tar.gz && tar -xf vggface2_test.tar.gz')
        os.system(f'cd {os.path.join(self.dir, "meta")} && tar -xf bb_landmark.tar.gz')
        os.system('touch DATASET_IS_READY')
      else:
        # 数据集存在，创建标记
        if not os.path.exists('DATASET_IS_READY'):
          os.system('touch DATASET_IS_READY')     
    else:
      while True:
        # 等待直到存在指定文件
        if os.path.exists('DATASET_IS_READY'):
          break
        time.sleep(5)

    meta_file_name = 'loose_bb_train.csv' if self.train_or_test == 'train' else 'loose_bb_test.csv'
    
    self.data_list = []
    person_id_map = {}
    with open(os.path.join(self.dir, 'meta', 'bb_landmark', meta_file_name), 'r') as fp:
      # skip first line
      content = fp.readline()
      # file_name, x,y,w,h
      content = fp.readline()
      content = content.strip()
      while content:
        file_name, x,y,w,h = content.split(',')
        file_name = file_name[1:-1]
        person_id, _ = file_name.split('/')
        if person_id not in person_id_map:
          person_id_map[person_id] = len(person_id_map)
        
        self.data_list.append({
          'filepath': os.path.join(self.dir, 'data', self.train_or_test, f'{file_name}.jpg'),
          'person_id': person_id_map[person_id],
          'bbox': [int(x), int(y), int(x)+int(w), int(y)+int(h)]
        })
              
        content = fp.readline()
        content = content.strip()


    if self.train_or_test != 'train':
      # 重新设置person_id_map
      person_id_map = {}
      with open(os.path.join(self.dir, 'meta', 'bb_landmark', 'loose_bb_train.csv'), 'r') as fp:
        # skip first line
        content = fp.readline()
        # file_name, x,y,w,h
        content = fp.readline()
        content = content.strip()
        while content:
          file_name, x,y,w,h = content.split(',')
          file_name = file_name[1:-1]
          person_id, _ = file_name.split('/')
          if person_id not in person_id_map:
            person_id_map[person_id] = len(person_id_map)
                
          content = fp.readline()
          content = content.strip()

    print(f'person id num {len(person_id_map)}')
    print(f'sample num {len(self.data_list)}')
    
  @property
  def size(self):
    return len(self.data_list)

  def at(self, id):
    info = self.data_list[id]
    image = cv2.imread(info['filepath'])
    h,w = image.shape[:2]
    x0,y0,x1,y1 = info['bbox']
    x0 = max(0, x0)
    y0 = max(0, y0)
    x1 = min(w, x1)
    y1 = min(h, y1)
    id = info['person_id']
    crop_image = image[y0:y1,x0:x1].copy()
    return crop_image, {'label':id, 'image_meta': {'image_shape': crop_image.shape[:2]}}

# vgg = VGGFace('test', '/opt/tiger/handdetJ/dataset/vggface2')
# size = vgg.size
# print(f'vgg size {size}')
# for i in range(size):
#   data = vgg.sample(i)
#   cv2.imwrite('./aabb.png', data['image'])
#   print(i)


import os.path as osp

import torch.distributed as dist
from antgo.framework.helper.parallel import is_module_wrapper
from .evaluation import DistEvalHook, EvalHook
from torch.nn.modules.batchnorm import _BatchNorm
from .logger import LoggerHook
import warnings

class SubModulesEvalHook(EvalHook):
    def __init__(self, *args, evaluated_modules=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.evaluated_modules = evaluated_modules

    def before_run(self, runner):
        if is_module_wrapper(runner.model):
            model = runner.model.module
        else:
            model = runner.model
        assert hasattr(model, "submodules")
        assert hasattr(model, "inference_on")

    def after_train_iter(self, runner):
        """Called after every training iter to evaluate the results."""
        if not self.by_epoch and self._should_evaluate(runner):
            for hook in runner._hooks:
                if isinstance(hook, LoggerHook):
                    hook.after_train_iter(runner)
            runner.log_buffer.clear()

            self._do_evaluate(runner)

    def _do_evaluate(self, runner):
        """perform evaluation and save ckpt."""
        # Synchronization of BatchNorm's buffer (running_mean
        # and running_var) is not supported in the DDP of pytorch,
        # which may cause the inconsistent performance of models in
        # different ranks, so we broadcast BatchNorm's buffers
        # of rank 0 to other ranks to avoid this.
        if is_module_wrapper(runner.model):
            model_ref = runner.model.module
        else:
            model_ref = runner.model
        if not self.evaluated_modules:
            submodules = model_ref.submodules
        else:
            submodules = self.evaluated_modules
        key_scores = []

        for submodule in submodules:
            # change inference on
            model_ref.inference_on = submodule
            results = self.test_fn(runner.model, self.dataloader)
            runner.log_buffer.output['eval_iter_num'] = len(self.dataloader)
            key_score = self.evaluate(runner, results, prefix=submodule)
            key_scores.append(key_score)

        if len(key_scores) == 0:
            key_scores = [None]
        best_score = key_scores[0]
        for key_score in key_scores:
            if hasattr(self, "compare_func") and self.compare_func(
                key_score, best_score
            ):
                best_score = key_score

        # runner.log_buffer.output["eval_iter_num"] = len(self.dataloader)
        if self.save_best:
            self._save_ckpt(runner, best_score)

    def evaluate(self, runner, results, prefix=""):
        """Evaluate the results.

        Args:
            runner (:obj:`mmcv.Runner`): The underlined training runner.
            results (list): Output results.
        """
        eval_res = {}
        if self.metric_func is None:
            eval_res = self.dataloader.dataset.evaluate(
                results, logger=runner.logger, **self.eval_kwargs)
        else:
            gts = []
            for gt_i in range(len(self.dataloader.dataset)):
                gts.append(self.dataloader.dataset.get_ann_info(gt_i))

            eval_res = self.metric_func(results, gts)

        for name, val in eval_res.items():
            runner.log_buffer.output[(".").join([prefix, name])] = val
        runner.log_buffer.ready = True

        if self.save_best is not None:
            # If the performance of model is pool, the `eval_res` may be an
            # empty dict and it will raise exception when `self.save_best` is
            # not None. More details at
            # https://github.com/open-mmlab/mmdetection/issues/6265.
            if not eval_res:
                warnings.warn(
                    'Since `eval_res` is an empty dict, the behavior to save '
                    'the best checkpoint will be skipped in this evaluation.')
                return None

            if self.key_indicator == "auto":
                # infer from eval_results
                self._init_rule(self.rule, list(eval_res.keys())[0])
            return eval_res[self.key_indicator]

        return None


class SubModulesDistEvalHook(DistEvalHook):
    def __init__(self, *args, evaluated_modules=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.evaluated_modules = evaluated_modules

    def before_run(self, runner):
        if is_module_wrapper(runner.model):
            model = runner.model.module
        else:
            model = runner.model
        assert hasattr(model, "submodules")
        assert hasattr(model, "inference_on")

    def after_train_iter(self, runner):
        """Called after every training iter to evaluate the results."""
        if not self.by_epoch and self._should_evaluate(runner):
            for hook in runner._hooks:
                if isinstance(hook, LoggerHook):
                    hook.after_train_iter(runner)
            runner.log_buffer.clear()

            self._do_evaluate(runner)

    def _do_evaluate(self, runner):
        """perform evaluation and save ckpt."""
        # Synchronization of BatchNorm's buffer (running_mean
        # and running_var) is not supported in the DDP of pytorch,
        # which may cause the inconsistent performance of models in
        # different ranks, so we broadcast BatchNorm's buffers
        # of rank 0 to other ranks to avoid this.

        if self.broadcast_bn_buffer:
            model = runner.model
            for name, module in model.named_modules():
                if isinstance(module, _BatchNorm) and module.track_running_stats:
                    dist.broadcast(module.running_var, 0)
                    dist.broadcast(module.running_mean, 0)

        tmpdir = self.tmpdir
        if tmpdir is None:
            tmpdir = osp.join(runner.work_dir, ".eval_hook")

        if is_module_wrapper(runner.model):
            model_ref = runner.model.module
        else:
            model_ref = runner.model
        if not self.evaluated_modules:
            submodules = model_ref.submodules
        else:
            submodules = self.evaluated_modules
        key_scores = []

        for submodule in submodules:
            # change inference on
            model_ref.inference_on = submodule
            results = self.test_fn(
                runner.model,
                self.dataloader,
                tmpdir=tmpdir,
                gpu_collect=False,
            )
            if runner.rank == 0:
                key_score = self.evaluate(runner, results, prefix=submodule)
                if key_score is not None:
                    key_scores.append(key_score)

        if runner.rank == 0:
            runner.log_buffer.ready = True
            if len(key_scores) == 0:
                key_scores = [None]
            best_score = key_scores[0]
            for key_score in key_scores:
                if hasattr(self, "compare_func") and self.compare_func(
                    key_score, best_score
                ):
                    best_score = key_score

            # runner.log_buffer.output["eval_iter_num"] = len(self.dataloader)
            if self.save_best:
                self._save_ckpt(runner, best_score)

    def evaluate(self, runner, results, prefix=""):
        """Evaluate the results.

        Args:
            runner (:obj:`mmcv.Runner`): The underlined training runner.
            results (list): Output results.
        """
        eval_res = {}
        if self.metric_func is None:
            eval_res = self.dataloader.dataset.evaluate(
                results, logger=runner.logger, **self.eval_kwargs)
        else:
            gts = []
            for gt_i in range(len(self.dataloader.dataset)):
                gts.append(self.dataloader.dataset.get_ann_info(gt_i))

            eval_res = self.metric_func(results, gts)

        for name, val in eval_res.items():
            runner.log_buffer.output[(".").join([prefix, name])] = val

        if self.save_best is not None:
            # If the performance of model is pool, the `eval_res` may be an
            # empty dict and it will raise exception when `self.save_best` is
            # not None. More details at
            # https://github.com/open-mmlab/mmdetection/issues/6265.
            if not eval_res:
                warnings.warn(
                    'Since `eval_res` is an empty dict, the behavior to save '
                    'the best checkpoint will be skipped in this evaluation.')
                return None

            if self.key_indicator == "auto":
                # infer from eval_results
                self._init_rule(self.rule, list(eval_res.keys())[0])
            return eval_res[self.key_indicator]

        return None



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/6 23:17
# @File    : __init__.py.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function


from pathlib import Path
from .operator_registry import OperatorRegistry # pylint: disable=import-outside-toplevel


register = OperatorRegistry.register
resolve = OperatorRegistry.resolve


DEFAULT_LOCAL_CACHE_ROOT = Path.home() / '.antgo'
LOCAL_PIPELINE_CACHE = DEFAULT_LOCAL_CACHE_ROOT / 'pipelines'
LOCAL_OPERATOR_CACHE = DEFAULT_LOCAL_CACHE_ROOT / 'operators'



import warnings

import torch
import torch.nn.functional as F
from torch import nn

from antgo.framework.helper.models.classification.losses.accuracy import Accuracy
from antgo.framework.helper.models.classification.losses.cross_entropy_loss import *
from ...builder import *


@HEADS.register_module()
class ClsHead(nn.Module):
    """classification head.

    Args:
        loss (dict): Config of classification loss.
        topk (int | tuple): Top-k accuracy.
        cal_acc (bool): Whether to calculate accuracy during training.
            If you use Mixup/CutMix or something like that during training,
            it is not reasonable to calculate accuracy. Defaults to False.
    """

    def __init__(self,
                 loss=dict(type='CrossEntropyLoss', loss_weight=1.0, class_weight=None),
                 init_cfg=dict(use_softmax=True)):
        super().__init__()

        assert isinstance(loss, dict)
        self.compute_loss = CrossEntropyLoss(loss_weight=loss['loss_weight'], class_weight=loss['class_weight'])
        self.use_softmax = init_cfg.get('use_softmax', True)

    def loss(self, cls_score, gt_label, **kwargs):
        num_samples = len(cls_score)
        losses = dict()
        # compute loss
        loss = self.compute_loss(
            cls_score, gt_label, avg_factor=num_samples, **kwargs)
        losses['loss'] = loss
        return losses

    def forward_train(self, cls_score, gt_label, **kwargs):
        if isinstance(cls_score, tuple):
            cls_score = cls_score[-1]
            
        losses = self.loss(cls_score, gt_label.view(-1), **kwargs)
        return losses

    def simple_test(self, x, **kwargs):
        """Inference without augmentation.

    Args:
            x (tuple[Tensor]): The input features.
                Multi-stage inputs are acceptable but only the last stage will
                be used to classify. The shape of every item should be
                ``(num_samples, in_channels)``.
            softmax (bool): Whether to softmax the classification score.
            post_process (bool): Whether to do post processing the
                inference results. It will convert the output to a list.

        Returns:
            Tensor | list: The inference results.

                - If no post processing, the output is a tensor with shape
                  ``(num_samples, num_classes)``.
                - If post processing, the output is a multi-dimentional list of
                  float and the dimensions are ``(num_samples, num_classes)``.
        """
        cls_score = self.pre_logits(x)

        if self.use_softmax:
            pred = (
                F.softmax(cls_score, dim=1) if cls_score is not None else None)
        else:
            pred = cls_score
        return pred

    def pre_logits(self, x):
        if isinstance(x, tuple):
            x = x[-1]
        return x



from antgo.framework.helper.runner.builder import *
import numpy as np


@MEASURES.register_module()
class AccuracyEval(object):
    def __init__(self, topk=(1, ), thrs=0.) -> None:
        self.topk = topk
        self.thrs = thrs

    def keys(self):
        # 约束使用此评估方法，需要具体的关键字信息
        return {'pred': ['pred'], 'gt': ['label']}

    def __call__(self, preds, gts):
        preds_reformat = [pred['pred'] for pred in preds]
        gts_reformat = [gt['label'] for gt in gts]
        preds = np.stack(preds_reformat, 0)
        gts = np.array(gts_reformat).astype(np.int32)

        maxk = max(self.topk)
        num = preds.shape[0]

        static_inds = np.indices((num, maxk))[0]
        pred_label = preds.argpartition(-maxk, axis=1)[:, -maxk:]
        pred_score = preds[static_inds, pred_label]

        sort_inds = np.argsort(pred_score, axis=1)[:, ::-1]
        pred_label = pred_label[static_inds, sort_inds]
        pred_score = pred_score[static_inds, sort_inds]

        eval_values = {}
        for k in self.topk:
            correct_k = pred_label[:, :k] == gts.reshape(-1, 1)
            # Only prediction values larger than thr are counted as correct
            _correct_k = correct_k & (pred_score[:, :k] > self.thrs)
            _correct_k = np.logical_or.reduce(_correct_k, axis=1)
            eval_values[f'top_{k}'] = _correct_k.sum() * 100. / num

        return eval_values


"""A `Module` implement the `AntgoFunc` class."""
__all__ = ['AntgoFunc', 'bind']


import ctypes
import hashlib
import warnings
import numpy as np
from . import glue
from .internal.dtype import DType, CStruct, TemplateType, UnknownCType
from .building.build_utils import config


def get_func_idcode(func_name, arg_types):
    """Get Function IDCode

    Parameters
    ----------
    func_name: str
        the name of function
    arg_types: list of DType

    Returns
    -------
    idcode: str
        IDCode
    """
    arg_types_str = ','.join([e.cname for e in arg_types])
    idcode = '{func_name}:{arg_types_str}'.format(
        func_name=func_name,
        arg_types_str=arg_types_str)
    return idcode


def get_idcode_hash(idcode):
    """Get the hash string of IDCode

    Parameters
    ----------
    idcode: str
    arg_types: list of DType

    Returns
    -------
    Hash String of IDCode: str
    """
    idcode_sp = idcode.split(':')
    func_name = idcode_sp[0]
    md5 = hashlib.md5()
    md5.update(idcode[len(func_name) + 1:].encode('utf-8'))
    return '{}_{}'.format(func_name, md5.hexdigest()[:8])


class CFuncTensor:
    def __init__(self, var, ptype, glue_mod):
        self.var = var
        self.ptype = ptype
        self.glue_mod = glue_mod

    @property
    def is_const(self):
        return self.ptype.is_const


class CStructArg:
    def __init__(self, var, ptype):
        self.var = var
        self.ptype = ptype

    @property
    def is_const(self):
        return self.ptype.is_const

def _wait_to_read(var):
    if hasattr(var, 'wait_to_read'):
        var.wait_to_read()


def _wait_to_write(var):
    if hasattr(var, 'wait_to_write'):
        var.wait_to_write()


def _get_raw_pointer(arg, const_vars, mutable_vars):
    if isinstance(arg, CFuncTensor):
        p = arg.glue_mod.Tensor(arg.var).data_ptr
        if isinstance(p, (list, tuple)):
            p, v = p
            if arg.is_const:
                const_vars.append(v)
            else:
                mutable_vars.append((arg.var, v))
        return p
    if isinstance(arg, CStructArg):
        if arg.is_const:
            const_vars.append(arg.var)
            return ctypes.byref(const_vars[-1])
        else:
            mutable_vars.append((arg.var.val, arg.var))
            return ctypes.byref(mutable_vars[-1][-1])

    return arg


def _get_async_pointer(arg):
    if isinstance(arg, CFuncTensor):
        return arg.glue_mod.Tensor(arg.var).async_data_ptr
    return arg


def _get_raw_pointers(args, const_vars, mutable_vars):
    return [_get_raw_pointer(a, const_vars, mutable_vars) for a in args]


def _get_async_pointers(args):
    return list(map(_get_async_pointer, args))


def _arg_wait_to_rw(arg):
    if isinstance(arg, CFuncTensor):
        if arg.is_const:
            _wait_to_read(arg)
        else:
            _wait_to_write(arg)


def _args_wait_to_rw(args):
    for arg in args:
        _arg_wait_to_rw(arg)


class CFuncDef:
    """The definition of CFunction."""
    KERNEL = 1
    FUNC = 2

    def __init__(self, func_name, func_kind, arg_names=None, arg_types=None, rtn_type=None,
                 template_list=None, loader=None, loader_kwargs=None):
        self.func_name = func_name
        self.func_kind = func_kind
        self.arg_names = arg_names or list()
        self.arg_types = arg_types
        self.rtn_type = rtn_type
        self.template_list = template_list or list()
        self.loader = loader
        self.loader_kwargs = loader_kwargs

    def __call__(self, arg_datas, arg_types, dev_id, glue_mod=None, using_async=False):
        if dev_id is None:
            ctx = 'cpu'
            dev_id = -1
        else:
            ctx = config.GPU_BACKEND
        # function loader
        func = self.loader(self, arg_types, ctx, **self.loader_kwargs)
        if using_async and glue_mod is not None:
            async_func = func.get_async_func(glue_mod)
            if async_func is not None:
                return async_func(*_get_async_pointers(arg_datas))

        _args_wait_to_rw(arg_datas)
        const_vars = []
        mutable_vars = []
        raw_pointers = _get_raw_pointers(arg_datas, const_vars, mutable_vars)
        res = None
        if self.func_kind == CFuncDef.KERNEL:
            res = func(dev_id, *raw_pointers)
        elif self.func_kind == CFuncDef.FUNC:
            res = func(*raw_pointers)
        else:
            raise TypeError(
                'Unsupported func kind: {}'.format(self.func_kind))
        
        out = [] if res is None else [res]
        for target, value in mutable_vars:
            if isinstance(value, np.ndarray):
                # for CFuncTensor
                target[:] = value
                if res is None:
                    out.append(target)
            else:
                # for CStructArg
                target = np.ctypeslib.as_array(value.data, shape=[value.dims[i] for i in range(value.dim_size)])
                # 自动释放
                if value.data != mutable_vars[0][1].data:
                    # clone 
                    target = target.copy()

                    # c函数内部创建，需要销毁内存
                    func.clear(ctypes.byref(value))

                if res is None:
                    out.append(target)

        return out


class AntgoFunc:
    """An encapsulation for CFunction

    Parameters:
    -----------
    name: str
        function name
    func: CFuncDef
    """

    def __init__(self, name, func):
        self.name = name
        self.func = func

        self.wait_to_read_list = []
        self.wait_to_write_list = []
        for i, ptype in enumerate(self.func.arg_types):
            if ptype.is_pointer:
                if ptype.is_const:
                    self.wait_to_read_list.append(i)
                else:
                    self.wait_to_write_list.append(i)

    def __call__(self, *args, **kwargs):
        # move kwargs into args
        args = list(args)
        for name in self.func.arg_names[len(args):]:
            args.append(kwargs[name])

        # type check
        arg_datas = []
        dev_id = None
        arg_types = []
        template_mapping = dict()

        glue_mod = self._get_glue_mod(args)
        using_async = config.USING_ASYNC_EXEC and glue_mod is not None and hasattr(
            glue_mod, 'get_async_func')

        if not using_async:
            # Pre-process
            for i in self.wait_to_read_list:
                _wait_to_read(args[i])
            for i in self.wait_to_write_list:
                _wait_to_write(args[i])

        try:
            for var, ptype in zip(args, self.func.arg_types):
                if ptype.is_pointer:
                    if hasattr(ptype, 'constructor'):
                        var_dev_id = None
                        ctype = ctypes.POINTER(ptype.cstruct)
                        try:
                            data = CStructArg(ptype.constructor(var), ptype)
                        except TypeError:
                            data = CStructArg(ptype.constructor(*var), ptype)
                    else:
                        # The type of `var` is Tensor.
                        data, var_dev_id, ctype = self._get_tensor_info(
                            var, ptype, template_mapping, using_async)
                else:
                    # The type of `var` is Scalar.
                    data, var_dev_id, ctype = self._get_scalar_info(var, ptype)

                arg_datas.append(data)
                if isinstance(ctype, UnknownCType):
                    ctype.is_const = ptype.is_const
                    arg_types.append(ctype)
                else:
                    # pointer
                    arg_types.append(DType(ctype, is_const=ptype.is_const))

                # update `dev_id`
                if var_dev_id is not None:
                    if dev_id is not None:
                        assert var_dev_id == dev_id, ValueError(
                            "Don't use multiple devices in a call :-(")
                    else:
                        dev_id = var_dev_id

            # try to know the unknown ctype
            for i, vtype in enumerate(arg_types):
                if isinstance(vtype, UnknownCType):
                    assert vtype.tname in template_mapping,\
                        Exception(
                            'Unknown template name: {}'.format(vtype.tname))
                    ctype = template_mapping[vtype.tname]._type_
                    arg_types[i] = DType(ctype, vtype.is_const)
                    arg_datas[i] = ctype(arg_datas[i])
        except TypeError:
            raise TypeError('Unmatched parameters list of the function `{}`:\n\t{}\n\t\tvs\n\t{}'.format(
                self.name, self.func.arg_types, list(map(type, args))))

        rtn = self.func(arg_datas=arg_datas,
                        arg_types=arg_types,
                        dev_id=dev_id,
                        glue_mod=glue_mod,
                        using_async=using_async)
        return rtn

    @staticmethod
    def _get_tensor_info(var, ptype, template_mapping, using_async=False):
        """Get tensor info

        Parameters
        ----------
        var: object
            input variable
        ptype: DType | TemplateType
            the type of argument
        template_mapping: dict
            the mapping from template name to ctype
        using_async: bool
            whether to use asynchronous execution

        Returns
        -------
        data: CFuncTensor
        dev_id: int | None
            the id of device
        ctype: ctypes.POINTER | ctypes.c_*
            the ctype of data
        """
        glue_mod = glue.backend.get_var_glue(var)
        if glue_mod is None:
            raise TypeError()
        data = CFuncTensor(var, ptype, glue_mod)
        tensor = glue_mod.Tensor(var)
        dev_id = tensor.dev_id
        ctype = ctypes.POINTER(tensor.ctype)
        if isinstance(ptype, DType):
            expected_ctype = ptype.ctype
        elif ptype.tname in template_mapping:
            expected_ctype = template_mapping[ptype.tname]
        else:
            template_mapping[ptype.tname] = expected_ctype = ctype
        assert ctype == expected_ctype,\
            TypeError('Expected Type {} instead of {}'.format(
                expected_ctype, ctype))
        return data, dev_id, ctype

    @staticmethod
    def _get_scalar_info(var, ptype):
        """Get scalar info

        Parameters
        ----------
        var: object
            input variable
        ptype: DType | TemplateType
            the type of argument

        Returns
        -------
        data: ctyoes.c_void_p
            the pointer of data
        dev_id: int | None
            the id of device
        ctype: ctypes.POINTER | ctypes.c_*
            the ctype of data
        """

        dev_id = None
        if isinstance(ptype, TemplateType):
            data = var
            ctype = type(var) if hasattr(
                var, '_type_') else UnknownCType(ptype.tname)
        else:
            data = var if isinstance(
                var, ctypes.c_void_p) else ptype.ctype(var)
            ctype = ptype.ctype
        return data, dev_id, ctype

    @staticmethod
    def _get_glue_mod(datas):
        mods = map(glue.backend.get_var_glue, datas)
        mods = list(filter(lambda x: x is not None, mods))
        if mods:
            glue_mod = mods[0]
            # all glue modules in datas are consistent
            if all(map(lambda x: x == glue_mod, mods)):
                return glue_mod
        return None

    def build(self, ctx, template_types=None):
        """Build this function

        Parameters
        ----------
        ctx: str
            context Name
        template_types: list or tuple or dict, default: []
            list: a list of template type Names
            tuple: a tuple of template type Names
            dict: a mapping from template name to type name

        Examples
        --------
        >>> mobula.func.add.build('cpu', ['float'])
        """
        arg_types = []
        par_type = self.func.arg_types
        if template_types is None:
            template_types = list()
        if isinstance(template_types, (list, tuple)):
            template_mapping = dict()  # tname -> ctype
            for vtype in par_type:
                if isinstance(vtype, TemplateType):
                    tname = vtype.tname
                    if tname in template_mapping:
                        ctype = template_mapping[tname]
                    else:
                        ctype = getattr(ctypes, 'c_{}'.format(
                            template_types.pop(0)))
                        template_mapping[tname] = ctype
                    arg_types.append(vtype(ctype))
                else:
                    arg_types.append(vtype)
            assert not template_types, Exception('redundant type')
        else:
            assert isinstance(template_types, dict), TypeError(
                'The type of template_types should be list or tuple or dict.')
            template_name = set()
            for vtype in par_type:
                if isinstance(vtype, TemplateType):
                    tname = vtype.tname
                    assert tname in template_types, KeyError(
                        'Unknown Template Type: {}'.format(tname))
                    template_name.add(tname)
                    ctype = getattr(ctypes, 'c_{}'.format(
                        template_types[tname]))
                    arg_types.append(vtype(ctype))
                else:
                    arg_types.append(vtype)
            assert len(template_name) == len(template_types), Exception(
                'Different template name: {} vs {}'.format(
                    template_name, set(template_types.keys())))
        func = self.func
        func.loader(func, arg_types, ctx, **func.loader_kwargs)


_binded_functions = dict()


def bind(functions):
    global _binded_functions
    """Bind Functions to mobula.func.<function name>

    Parameters
    ----------
    functions: dict
        name -> CFuncDef
    """
    for k, func in functions.items():
        if k in _binded_functions:
            warnings.warn('Duplicated function name {}'.format(k))
        func = AntgoFunc(k, func)
        globals()[k] = func
        _binded_functions[k] = func



import os
import logging
import json
from antgo.interactcontext import InteractContext

def browser_images(src_file, tags, white_users_str, feedback=True, user_input=False):
    # src check
    if not os.path.exists(src_file):
        logging.error(f"{src_file} not existed")
        return

    if src_file is None or src_file == '' or src_file == './':
         title = 'default'
         src_file = None
    else:
        if not src_file.endswith('.json'):
            logging.error(f"{src_file} must be json file")
            return
        title = src_file.split('/')[-1].split('.')[0]

    if tags is not None:
        tags = tags.split(',')
    else:
        tags = []
    white_users = {}
    if white_users_str is not None:
        for t in white_users_str.split(','):
            user_name, password = t.split(':')
            white_users.update({
                user_name: {'password': password}
            })
    ctx = InteractContext()
    ctx.browser.start(f"{title}", config = {
            'tags': tags,
            'user_input': user_input,
            'white_users': white_users,
        }, json_file=src_file)
    
    logging.info('Waiting data browser stop.')
    ctx.browser.waiting(not feedback)
    
    if feedback:
        content = ctx.browser.download()
        
        src_folder = os.path.dirname(src_file)
        src_name = src_file.split('/')[-1].split('.')[0]
        with open(os.path.join(src_folder, f'{src_name}_browser.json'), 'w') as fp:
            json.dump(content, fp)
    ctx.browser.exit()


from email.utils import localtime
import logging
import sys
import numpy as np
import torch
import torch.distributed as dist
from antgo.framework.helper.utils import build_from_cfg
from antgo.framework.helper.runner.dist_utils import get_dist_info
from tfrecord.reader import *
from tfrecord import iterator_utils
from antgo.framework.helper.dataset.builder import DATASETS
import copy
from antgo.dataflow.datasetio import *
from antgo.framework.helper.fileio.file_client import *
import threading
import json
from pprint import pprint


def _cycle(iterator_fn: typing.Callable) -> typing.Iterable[typing.Any]:
    """Create a repeating iterator from an iterator generator."""
    while True:
        for element in iterator_fn():
            yield element

def _sample_iterators(iterators, ratios, infinite, remain_sample_num):
    """Retrieve info generated from the iterator(s) according to their
    sampling ratios.

    Params:
    -------
    iterators: list of iterators
        All iterators (one for each file).

    ratios: list of int
        The ratios with which to sample each iterator.
    
    infinite: bool, optional, default=True
        Whether the returned iterator should be infinite or not

    Yields:
    -------
    item: Any
        Decoded bytes of features into its respective data types from
        an iterator (based off their sampling ratio).
    """
    ext_iterators = None
    iterator_num = len(iterators)
    if infinite:
        iterators = [_cycle(iterator) for iterator in iterators]
    else:
        ext_iterators = [_cycle(iterator) for iterator in iterators]
        iterators = [iterator() for iterator in iterators]
    
    ratios = np.array(ratios)
    ratios = ratios / ratios.sum()
        
    in_remain_sample_mode = False
    while iterators or in_remain_sample_mode:
        try:
            if not in_remain_sample_mode:
                choice = np.random.choice(len(ratios), p=ratios)
                yield next(iterators[choice])
            else:
                if remain_sample_num == 0:
                    in_remain_sample_mode = False
                    break
                remain_sample_num -= 1

                yield next(ext_iterators[np.random.randint(0, iterator_num)])
        except StopIteration:
            if iterators:
                del iterators[choice]
                ratios = np.delete(ratios, choice)
                ratios = ratios / ratios.sum()

                if len(iterators) == 0 and remain_sample_num > 0 and not in_remain_sample_mode:
                    in_remain_sample_mode = True


def _order_iterators(iterators):
    iterators = [iterator() for iterator in iterators]
    choice = 0
    while iterators:
        try:
            yield next(iterators[choice])
        except StopIteration:
            if iterators:
                del iterators[choice]
                choice += 1


@DATASETS.register_module()
class TFDataset(torch.utils.data.IterableDataset):
    """Parse (generic) TFRecords dataset into `IterableDataset` object,
    which contain `np.ndarrays`s. By default (when `sequence_description`
    is None), it treats the TFRecords as containing `tf.Example`.
    Otherwise, it assumes it is a `tf.SequenceExample`.

    Params:
    -------
    data_path_list: str
        The path to the tfrecords file.

    index_path: str or None
        The path to the index file.

    description: list or dict of str, optional, default=None
        List of keys or dict of (key, value) pairs to extract from each
        record. The keys represent the name of the features and the
        values ("byte", "float", or "int") correspond to the data type.
        If dtypes are provided, then they are verified against the
        inferred type for compatibility purposes. If None (default),
        then all features contained in the file are extracted.

    shuffle_queue_size: int, optional, default=None
        Length of buffer. Determines how many records are queued to
        sample from.

    transform : a callable, default = None
        A function that takes in the input `features` i.e the dict
        provided in the description, transforms it and returns a
        desirable output.

    sequence_description: list or dict of str, optional, default=None
        Similar to `description`, but refers to the sequence features
        within a `SequenceExample`. When this field is `None`, then it
        is assumed that an `Example` is being read otherwise, a
        `SequenceExample` is read. If an empty list or dictionary is
        passed, then all features contained in the file are extracted.

    compression_type: str, optional, default=None
        The type of compression used for the tfrecord. Choose either
        'gzip' or None.

    """
    def __init__(self,
                 data_folder,
                 ratios: typing.Union[typing.List[float], None]=None,
                 description: typing.Union[typing.List[str], typing.Dict[str, str], None] = None,
                 shuffle_queue_size: typing.Optional[int] = 1024,
                 pipeline: typing.Optional[typing.List]=None, 
                 weak_pipeline: typing.Optional[typing.List]=None, 
                 strong_pipeline: typing.Optional[typing.List]=None, 
                 sequence_description: typing.Union[typing.List[str], typing.Dict[str, str], None] = None,
                 compression_type: typing.Optional[str] = None,
                 infinite: typing.Optional[bool] = False,
                 inputs_def=None,
                 sample_num_equalizer=True,
                 auto_ext_info=[]) -> None:
        super().__init__()
        if data_folder is None:
            # 不允许data_path_list和data_folder同时不设置
            logging.error('Must set data_folder')
            return

        # 自动下载指定数据(仅在首次运行时，进行下载)
        if ':///' in data_folder:
            terms = data_folder.split('/')
            dataset_name = terms[-1]
            if terms[-1] == '*' or terms[-1] == '':
                dataset_name = terms[-2]

            # 远程存储，自动下载
            local_temp_folder = f'./temp_{dataset_name}'
            if os.environ.get('LOCAL_RANK', 0) == 0:
                if not os.path.exists(local_temp_folder):
                    # 数据集不存在，需要重新下载，并创建标记
                    # 创建临时目录
                    os.makedirs(local_temp_folder)

                    # 下载
                    file_client_get(data_folder, local_temp_folder)
                    os.system('touch DATASET_IS_READY')
                else:
                    # 数据集存在，创建标记
                    if not os.path.exists('DATASET_IS_READY'):
                        os.system('touch DATASET_IS_READY')                    
            else:
                while True:
                    # 等待直到存在指定文件
                    if os.path.exists('DATASET_IS_READY'):
                        break
                    time.sleep(5)

        # 替换为本地路径
        if ':///' in data_folder:
            terms = data_folder.split('/')
            dataset_name = terms[-1]
            if terms[-1] == '*' or terms[-1] == '':
                dataset_name = terms[-2]

            # 使用本地地址替换
            data_folder = f'./temp_{dataset_name}'

        # 遍历文件夹，发现所有tfrecord数据
        self.data_path_list = []
        for tfrecord_file in os.listdir(data_folder):
            if tfrecord_file.endswith('tfrecord'):
                tfrecord_file = '-'.join(tfrecord_file.split('/')[-1].split('-')[:-1]+['tfrecord'])
                self.data_path_list.append(f'{data_folder}/{tfrecord_file}')

        self.index_path_list = []
        for i in range(len(self.data_path_list)):
            tfrecord_file = self.data_path_list[i]
            folder = os.path.dirname(tfrecord_file)
            if tfrecord_file.endswith('tfrecord') or tfrecord_file.endswith('index'):
                index_file = '-'.join(tfrecord_file.split('/')[-1].split('-')[:-1]+['index'])
                index_file = f'{folder}/{index_file}'
                tfrecord_file = '-'.join(tfrecord_file.split('/')[-1].split('-')[:-1]+['tfrecord'])
                self.data_path_list[i] = f'{folder}/{tfrecord_file}'
            else:
                index_file = tfrecord_file+'-index'
                self.data_path_list[i] = tfrecord_file+'-tfrecord'
            self.index_path_list.append(index_file)

        if description is None:
            description = {}
            print('Using default tfrecord description.')
            tfdataset_file_path = os.path.realpath(__file__)
            parent_folder = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(tfdataset_file_path))))
            with open(os.path.join(parent_folder, 'resource', 'templates', 'sample_gt.json'), 'r') as fp:
                key_map_info = json.load(fp)

            for k in key_map_info.keys():                
                # 对于list数据直接转换为numpy
                if isinstance(key_map_info[k], list):
                    description[k] = 'numpy'
                # 对于dict数据转换成字符串
                elif isinstance(key_map_info[k], dict):
                    description[k] = 'dict'
                # 对bool数据转换成int
                elif isinstance(key_map_info[k], bool):
                    description[k] = 'int'
                elif isinstance(key_map_info[k], int):
                    description[k] = 'int'
                elif isinstance(key_map_info[k], float):
                    description[k] = 'float'
                elif isinstance(key_map_info[k], str):
                    description[k] = 'str'
                elif isinstance(key_map_info[k], bytes):
                    description[k] = 'byte'
                else:
                    print('unkown data type')

            # 默认字段，用于存储图像
            description['image'] = 'byte'

        self.description = {}
        self.raw_description = description
        for k, v in description.items():
            if v == 'numpy':
                self.description.update({
                    k: 'byte',
                    f'__{k}_type': 'int',
                    f'__{k}_shape': 'int'
                })
            elif v == 'str':
                self.description.update({
                    k: 'byte'
                })
            elif v == 'dict':
                self.description.update({
                    k: 'byte'
                })
            else:
                self.description.update({
                    k: v
                })

        self.sequence_description = sequence_description
        self.shuffle_queue_size = shuffle_queue_size
        self.sample_num_equalizer = sample_num_equalizer
        self.compression_type = compression_type
        self.ratios = ratios
        if shuffle_queue_size > 0:
            # 如果已经设置了shuffle queue，则设置默认ratios
            if self.ratios is None:
                self.ratios = [1 for _ in range(len(self.data_path_list))]
        self._fields = copy.deepcopy(inputs_def['fields']) if inputs_def else None
        self._alias = None
        if self._fields is not None and 'alias' in inputs_def:
            self._alias = copy.deepcopy(inputs_def['alias'])

        if self._fields is not None:
            if self._alias is None:
                self._alias = copy.deepcopy(self._fields)

        self.pipeline = []
        self.weak_pipeline = []
        self.strong_pipeline = []
        if pipeline is not None:
            from antgo.framework.helper.dataset import PIPELINES
            for transform in pipeline:
                if isinstance(transform, dict):
                    transform = build_from_cfg(transform, PIPELINES)
                    self.pipeline.append(transform)
                else:
                    raise TypeError('pipeline must be a dict')

            if weak_pipeline is not None and strong_pipeline is not None:
                for transform in weak_pipeline:
                    if isinstance(transform, dict):
                        transform = build_from_cfg(transform, PIPELINES)
                        self.weak_pipeline.append(transform)
                    else:
                        raise TypeError('weak_pipeline must be a dict')
                
                for transform in strong_pipeline:
                    if isinstance(transform, dict):
                        transform = build_from_cfg(transform, PIPELINES)
                        self.strong_pipeline.append(transform)
                    else:
                        raise TypeError('strong_pipeline must be a dict')
        self.infinite = infinite

        # 自动扩展样本信息
        self.auto_ext_info = auto_ext_info
        self.epoch = 10   
        self.sample_id = 0      # 先默认为0

        self.num_samples_list = []
        self.num_samples = 0
        for i, index_path in enumerate(self.index_path_list):
            index = np.loadtxt(index_path, dtype=np.int64)[:, 0]
            self.num_samples += len(index) 
            self.num_samples_list.append(len(index))

        rank, world_size = get_dist_info()
        self.rank = rank
        self.world_size = world_size

        self.select_index_list_in_world = []    # format: [[],[],[],...]
        if world_size > 1:
            # TODO，现在多卡实现基于文件级别的拆分，粒度较粗
            assert(len(self.data_path_list) >= world_size)

            # 公平选择，尽量确保每张卡有相同的样本数量
            self.select_index_list_in_world = self._fair_select(self.num_samples_list, world_size)

            # 每张卡期望使用的样本数（以最大为准，不足的通过后续重复采样补充）
            self.num_samples = 0
            for rank_i in range(world_size):
                num = np.sum([self.num_samples_list[i] for i in self.select_index_list_in_world[rank_i]])
                if self.num_samples < num:
                    self.num_samples = num

    def _select_next_set(self, num_samples_list, target_num):
        data = num_samples_list
        dp = [[0 for i in range(target_num+1)] for _ in range(len(data)+1)]

        for i in range(1,len(data)+1):
            for j in range(0, target_num+1):
                if j >= data[i-1]:
                    dp[i][j] = max(dp[i-1][j], dp[i-1][j-data[i-1]]+data[i-1])
                    pass
                else:
                    dp[i][j] = dp[i-1][j]
        
        select_list = []
        for i in range(len(data), 0, -1):
            if dp[i][j] > dp[i-1][j]:
                select_list.append(i-1)
                j -= data[i-1]

        return select_list
    
    def _fair_select(self, num_samples_list, world_size):
        select_index_list_in_world = []

        index_list = list(range(len(num_samples_list)))     # 维持全局索引
        num_samples_list = copy.deepcopy(num_samples_list)
        
        for i in range(world_size):
            if i == world_size - 1:
                select_index_list_in_world.append(index_list)
                break

            # 每次进行平均分配时，重新计算平均分配目标
            target_num = (sum(num_samples_list) + world_size-i-1)  // (world_size-i)
            select_local_index = self._select_next_set(num_samples_list, target_num)

            # 如果发现，本次分配方案，会导致剩余样本无法至少每个slot一个样本的话需要强制减少本次分配
            if len(index_list) - len(select_local_index) < world_size-i-1:
                remain_select_num = len(index_list) - (world_size-i-1)
                select_local_index = select_local_index[:remain_select_num]

            select_index = [index_list[j] for j in select_local_index]            
            select_index_list_in_world.append(select_index)

            num_samples_list = [num_samples_list[j] for j in range(len(num_samples_list)) if j not in select_local_index]               
            index_list = [k for k in index_list if k not in select_index]        

        return select_index_list_in_world

    def _arrange(self, sample, fields, alias):
        if fields is None:
            return sample      
          
        if type(fields[0]) == list or type(fields[0]) == tuple:
            warp_ins = []
            for alia, field in zip(alias, fields):
                one_ins = {}
                for aa, ff in zip(alia, field):
                    one_ins[aa] = sample[ff]
                
                warp_ins.append(one_ins)
            return warp_ins
        
        warp_ins = {}
        for alia, field in zip(alias, fields):
            warp_ins[alia] = sample[field]

        return warp_ins

    def __transform(self, sample):
        # 转换样本
        new_sample = {}
        for k in sample.keys():
            if not k.startswith('__'):
                if self.raw_description[k] == 'numpy':
                    dtype = numpy_dtype_map[sample[f'__{k}_type'][0]]
                    shape = tuple(sample[f'__{k}_shape'])
                    if isinstance(sample[k], bytes):
                        new_sample[k] = np.frombuffer(bytearray(sample[k]), dtype=dtype).reshape(shape).copy()
                    else:
                        new_sample[k] = np.frombuffer(bytearray(sample[k].tobytes()), dtype=dtype).reshape(shape).copy()
                elif self.raw_description[k] == 'str':
                    new_sample[k] = sample[k].tobytes().decode('utf-8')
                elif self.raw_description[k] == 'dict':
                    new_sample[k] = json.loads(sample[k].tobytes().decode('utf-8'))
                else:
                    new_sample[k] = sample[k]
        sample = new_sample
        weak_sample = None
        strong_sample = None
        if len(self.weak_pipeline) > 0 or len(self.strong_pipeline) > 0:
            # 扩展样本信息
            for ext_key in self.auto_ext_info:
                if ext_key not in sample:
                    sample[ext_key] = f'{self.sample_id}'

            weak_sample = copy.deepcopy(sample)
            for transform in self.weak_pipeline:
                weak_sample = transform(weak_sample)

            strong_sample = copy.deepcopy(weak_sample)
            for transform in self.strong_pipeline:
                strong_sample = transform(strong_sample)

        if weak_sample is not None and strong_sample is not None:
            for transform in self.pipeline:
                weak_sample = transform(weak_sample)

            for transform in self.pipeline:
                strong_sample = transform(strong_sample)

            weak_sample = self._arrange(weak_sample, self._fields, self._alias)
            strong_sample = self._arrange(strong_sample, self._fields, self._alias)
            return [weak_sample, strong_sample]
        else:
            # 扩展样本信息
            for ext_key in self.auto_ext_info:
                if ext_key not in sample:
                    sample[ext_key] = f'{self.sample_id}'

            for transform in self.pipeline:
                sample = transform(sample)

            # arange warp       
            sample = self._arrange(sample, self._fields, self._alias)        
            return sample

    def __iter__(self):
        # 获得线程信息
        worker_info = torch.utils.data.get_worker_info()

        # 每个epoch后，会重新调用__iter__
        # WARN: 对于多卡训练环境，在每次epoch时，重新编排数据文件的分配
        # 先按照单卡配置，进行赋初值
        real_num_samples = self.num_samples     # 单卡下期望的样本总数就是真实样本总数（只有多卡下才涉及补齐）
        ratios = self.ratios                    # 每个数据集的采样率
        data_path_list = self.data_path_list    # 总数据文件tfrecord列表
        index_path_list = self.index_path_list  # 总数据文件index列表
        if len(self.select_index_list_in_world) > 0:
            # 使用re_init_count进行shuffle （每个进程每个线程面对的是相同的re_init_count）
            local_select_index_list_in_world = copy.deepcopy(self.select_index_list_in_world)
            np.random.seed(self.epoch)  # 注意seed仅对当前线程有效，所以可以放心使用
            np.random.shuffle(local_select_index_list_in_world)

            # 选出当前卡，在本次epoch下使用的数据集
            # 每个线程，应该拥有相同的shuffle
            select_index_list = copy.deepcopy(local_select_index_list_in_world[self.rank])
            np.random.shuffle(select_index_list)
            
            # 基于选择的索引，获得具体样本数量列表
            use_data_path_num_list = [self.num_samples_list[i] for i in select_index_list]
            real_num_samples = np.sum(use_data_path_num_list)

            # 基于选择的索引，获得文件列表
            data_path_list = [self.data_path_list[i] for i in select_index_list]
            index_path_list = [self.index_path_list[i] for i in select_index_list]

            # 基于选择的索引，获得采样率列表
            if self.ratios is not None:
                ratios = [self.ratios[i] for i in select_index_list]

            # debug
            pprint(f'Rank {self.rank} (TOTAL NUM {real_num_samples} TARGET NUM {self.num_samples})')
            pprint(f'Rank {self.rank} local-index {local_select_index_list_in_world}')
            pprint(f'Rank {self.rank} thread id {worker_info.id} use index {select_index_list} and data path list {data_path_list}')

        remain_sample_num = 0
        if worker_info is not None:
            shard = worker_info.id, worker_info.num_workers
            np.random.seed(worker_info.seed % np.iinfo(np.uint32).max)
            if worker_info.num_workers != 1:
                expect_target_num = \
                    self.num_samples * (worker_info.id+1) // worker_info.num_workers - \
                    self.num_samples * (worker_info.id) // worker_info.num_workers

                real_target_num = real_num_samples * (worker_info.id+1) // worker_info.num_workers - \
                    real_num_samples * (worker_info.id) // worker_info.num_workers

                remain_sample_num = expect_target_num - real_target_num

            # debug
            pprint(f'Rank {self.rank} thread {worker_info.id} face remain sample_num {remain_sample_num}')
        else:
            shard = None

        if not self.sample_num_equalizer:
            remain_sample_num = 0

        loaders = [functools.partial(tfrecord_loader, data_path=data_path,
                                    index_path=index_path,
                                    shard=shard,
                                    description=self.description,
                                    sequence_description=self.sequence_description,
                                    compression_type=self.compression_type,
                                    )
                for data_path, index_path in zip(data_path_list, index_path_list)]

        it = None
        if ratios is not None and len(ratios) > 0:
            it = _sample_iterators(loaders, ratios, self.infinite, remain_sample_num)
        else:
            it = _order_iterators(loaders)

        if self.shuffle_queue_size:
            it = iterator_utils.shuffle_iterator(it, self.shuffle_queue_size)

        it = map(self.__transform, it)
        return it

    def __len__(self):
        return self.num_samples


# data = [2,4,6,7,8,5]
# data_sum = int(np.sum(data))

# dp = [[0 for i in range(data_sum//2+1)] for _ in range(len(data)+1)]

# for i in range(1,len(data)+1):
#     for j in range(0, data_sum//2+1):
#         if j >= data[i-1]:
#             dp[i][j] = max(dp[i-1][j], dp[i-1][j-data[i-1]]+data[i-1])
#             pass
#         else:
#             dp[i][j] = dp[i-1][j]

# j = data_sum//2
# print(j)

# print('select')
# for i in range(len(data), 0, -1):
#     if dp[i][j] > dp[i-1][j]:
#         print(i-1)
#         j -= data[i-1]

# print('sdf')

# abcd = TFDataset(data_path_list=[
#     '/root/workspace/dataset/hand-cls/yongchun_hand_gesture-00000-of-00003-tfrecord',
#     '/root/workspace/dataset/hand-cls/yongchun_hand_gesture-00001-of-00003-tfrecord',
#     '/root/workspace/dataset/hand-cls/yongchun_hand_gesture-00002-of-00003-tfrecord',
#     '/root/workspace/dataset/hand-cls/yongchun-cls-23-00000-of-00001-tfrecord']
# )
# print('sd')
# select_all = abcd._fair_select([2,4,6,7,8,5], 3)
# print(select_all)

# for ii in select_all:
#     print(np.sum(np.array([2,4,6,7,8,5])[ii]))